"""Helper functions for corpus parsing and loading"""
from __future__ import annotations

import datetime
import subprocess
import sys
import typing

import soundfile

from montreal_forced_aligner.data import FileExtensions, SoundFileInformation
from montreal_forced_aligner.exceptions import SoundFileError

SoundFileInfoDict = typing.Dict[str, typing.Union[int, float, str]]

supported_audio_extensions = {
    ".flac",
    ".ogg",
    ".aiff",
    ".mp3",
    ".opus",
    "flac",
    "ogg",
    "aiff",
    "mp3",
    "opus",
}

__all__ = ["load_text", "find_exts", "get_wav_info"]


def load_text(path: str) -> str:
    """
    Load a text file

    Parameters
    ----------
    path: str
        Text file to load

    Returns
    -------
    str
        Orthographic text of the file
    """
    with open(path, "r", encoding="utf8") as f:
        text = f.read().strip()
    return text


def find_exts(files: typing.List[str]) -> FileExtensions:
    """
    Find and group sound file extensions and transcription file extensions

    Parameters
    ----------
    files: List
        List of filename

    Returns
    -------
    :class:`~montreal_forced_aligner.data.FileExtensions`
        Data class for files found
    """
    exts = FileExtensions(set(), {}, {}, {}, {})
    for full_filename in files:
        try:
            filename, fext = full_filename.rsplit(".", maxsplit=1)
        except ValueError:
            continue
        fext = fext.lower()
        if fext == "wav":
            exts.wav_files[filename] = full_filename
        elif fext == "lab":
            exts.lab_files[filename] = full_filename
        elif (
            fext == "txt" and filename not in exts.lab_files
        ):  # .lab files have higher priority than .txt files
            exts.lab_files[filename] = full_filename
        elif fext == "textgrid":
            exts.textgrid_files[filename] = full_filename
        elif fext in supported_audio_extensions:
            exts.other_audio_files[filename] = full_filename
        exts.identifiers.add(filename)
    return exts


def get_wav_info(
    file_path: str, enforce_mono: bool = False, enforce_sample_rate: typing.Optional[int] = None
) -> SoundFileInformation:
    """
    Get sound file information

    Parameters
    ----------
    file_path: str
        Sound file path
    enforce_mono: bool
        Flag for whether to ensure that stereo files have the first channel extracted when processing
        them in Kaldi
    enforce_sample_rate: int, optional
        Sampling rate to enforce when sending data to Kaldi

    Returns
    -------
    :class:`~montreal_forced_aligner.data.SoundFileInformation`
        Sound information for format, duration, number of channels, bit depth, and
        sox_string for use in Kaldi feature extraction if necessary
    """
    _, format = file_path.rsplit(".", maxsplit=1)
    format = format.lower()
    num_channels = 0
    sample_rate = 0
    duration = 0
    sox_string = ""
    if format in {"mp3", "opus"}:
        if sys.platform != "win32" and format == "mp3":
            sox_proc = subprocess.Popen(
                ["soxi", f"{file_path}"], stderr=subprocess.PIPE, stdout=subprocess.PIPE, text=True
            )
            stdout, stderr = sox_proc.communicate()
            if stderr:
                raise SoundFileError(file_path, stderr)
            for line in stdout.splitlines():
                if line.startswith("Channels"):
                    num_channels = int(line.split(":")[-1].strip())
                elif line.startswith("Sample Rate"):
                    sample_rate = int(line.split(":")[-1].strip())
                elif line.startswith("Duration"):
                    duration_string = line.split(":", maxsplit=1)[-1].split("=")[0].strip()
                    duration = (
                        datetime.datetime.strptime(duration_string, "%H:%M:%S.%f")
                        - datetime.datetime(1900, 1, 1)
                    ).total_seconds()
                    break
            sample_rate_string = ""
            if enforce_sample_rate is not None:
                sample_rate_string = f" -r {enforce_sample_rate}"
            sox_string = f'sox "{file_path}" -t wav -b 16{sample_rate_string} - |'
        else:  # Fall back use ffmpeg if sox doesn't support the format
            ffmpeg_proc = subprocess.Popen(
                [
                    "ffprobe",
                    "-v",
                    "error",
                    "-hide_banner",
                    "-show_entries",
                    "stream=duration,channels,sample_rate",
                    "-of",
                    "default=noprint_wrappers=1",
                    "-i",
                    file_path,
                ],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            stdout, stderr = ffmpeg_proc.communicate()
            if stderr:
                raise SoundFileError(file_path, stderr)
            for line in stdout.splitlines():
                try:
                    key, value = line.strip().split("=")
                    if key == "duration":
                        duration = float(value)
                    elif key == "sample_rate":
                        sample_rate = int(value)
                    else:
                        num_channels = int(value)
                except ValueError:
                    pass
            mono_string = ""
            sample_rate_string = ""
            if num_channels > 1 and enforce_mono:
                mono_string = ' -af "pan=mono|FC=FL"'
            if enforce_sample_rate is not None:
                sample_rate_string = f" -ar {enforce_sample_rate}"
            sox_string = f'ffmpeg -nostdin -hide_banner -loglevel error -nostats -i "{file_path}" -acodec pcm_s16le -f wav{mono_string}{sample_rate_string} - |'
    else:
        use_sox = False
        with soundfile.SoundFile(file_path) as inf:
            frames = inf.frames
            sample_rate = inf.samplerate
            duration = frames / sample_rate
            num_channels = inf.channels
            try:
                bit_depth = int(inf.subtype.split("_")[-1])
                if bit_depth != 16:
                    use_sox = True
            except Exception:
                use_sox = True
        if format != "wav":
            use_sox = True
        if num_channels > 1 and enforce_mono:
            use_sox = True
        elif enforce_sample_rate is not None and sample_rate != enforce_sample_rate:
            use_sox = True
        sample_rate_string = ""
        if enforce_sample_rate is not None:
            sample_rate_string = f" -r {enforce_sample_rate}"
        if num_channels > 1 and enforce_mono:
            sox_string = f'sox "{file_path}" -t wav -b 16{sample_rate_string} - remix 1  |'
        elif use_sox:
            sox_string = f'sox "{file_path}" -t wav -b 16{sample_rate_string} - |'

    return SoundFileInformation(format, sample_rate, duration, num_channels, sox_string)



"""
Pronunciation dictionaries
==========================

"""

from montreal_forced_aligner.dictionary.mixins import DictionaryMixin, SanitizeFunction
from montreal_forced_aligner.dictionary.multispeaker import (
    MultispeakerDictionary,
    MultispeakerDictionaryMixin,
)
from montreal_forced_aligner.dictionary.pronunciation import (
    PronunciationDictionary,
    PronunciationDictionaryMixin,
)

__all__ = [
    "pronunciation",
    "multispeaker",
    "mixins",
    "DictionaryMixin",
    "SanitizeFunction",
    "MultispeakerDictionary",
    "MultispeakerDictionaryMixin",
    "PronunciationDictionary",
    "PronunciationDictionaryMixin",
]



"""Command line functions for segmenting audio files"""
from __future__ import annotations

import os
from typing import TYPE_CHECKING, List, Optional

from montreal_forced_aligner.exceptions import ArgumentError
from montreal_forced_aligner.segmenter import Segmenter

if TYPE_CHECKING:
    from argparse import Namespace


__all__ = ["create_segments", "validate_args", "run_create_segments"]


def create_segments(args: Namespace, unknown_args: Optional[List[str]] = None) -> None:
    """
    Run the sound file segmentation

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Command line arguments
    unknown_args: list[str]
        Optional arguments that will be passed to configuration objects
    """

    segmenter = Segmenter(
        corpus_directory=args.corpus_directory,
        temporary_directory=args.temporary_directory,
        **Segmenter.parse_parameters(args.config_path, args, unknown_args),
    )
    try:
        segmenter.segment()
        output_format = getattr(args, "output_format", None)
        segmenter.export_files(args.output_directory, output_format)
    except Exception:
        segmenter.dirty = True
        raise
    finally:
        segmenter.cleanup()


def validate_args(args: Namespace) -> None:
    """
    Validate the command line arguments

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments

    Raises
    ------
    :class:`~montreal_forced_aligner.exceptions.ArgumentError`
        If there is a problem with any arguments
    """
    args.output_directory = args.output_directory.rstrip("/").rstrip("\\")
    args.corpus_directory = args.corpus_directory.rstrip("/").rstrip("\\")
    if not os.path.exists(args.corpus_directory):
        raise ArgumentError(f"Could not find the corpus directory {args.corpus_directory}.")
    if not os.path.isdir(args.corpus_directory):
        raise ArgumentError(
            f"The specified corpus directory ({args.corpus_directory}) is not a directory."
        )

    if args.corpus_directory == args.output_directory:
        raise ArgumentError("Corpus directory and output directory cannot be the same folder.")


def run_create_segments(args: Namespace, unknown: Optional[List[str]] = None) -> None:
    """
    Wrapper function for running sound file segmentation

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments
    unknown: list[str]
        Parsed command line arguments to be passed to the configuration objects
    """
    validate_args(args)
    create_segments(args, unknown)



"""Module for ivector extractor training"""

from montreal_forced_aligner.ivector.trainer import (
    DubmTrainer,
    IvectorTrainer,
    TrainableIvectorExtractor,
)

__all__ = ["trainer", "DubmTrainer", "IvectorTrainer", "TrainableIvectorExtractor"]



"""Class definitions for LDA trainer"""
from __future__ import annotations

import multiprocessing as mp
import os
import re
import shutil
import subprocess
from queue import Empty
from typing import TYPE_CHECKING, Dict, List

import tqdm

from montreal_forced_aligner.abc import KaldiFunction
from montreal_forced_aligner.acoustic_modeling.triphone import TriphoneTrainer
from montreal_forced_aligner.data import MfaArguments
from montreal_forced_aligner.utils import (
    KaldiProcessWorker,
    Stopped,
    parse_logs,
    thirdparty_binary,
)

if TYPE_CHECKING:
    from montreal_forced_aligner.abc import MetaDict


__all__ = [
    "LdaTrainer",
    "CalcLdaMlltFunction",
    "CalcLdaMlltArguments",
    "LdaAccStatsFunction",
    "LdaAccStatsArguments",
]


class LdaAccStatsArguments(MfaArguments):
    """Arguments for :func:`~montreal_forced_aligner.acoustic_modeling.lda.LdaAccStatsFunction`"""

    dictionaries: List[str]
    feature_strings: Dict[str, str]
    ali_paths: Dict[str, str]
    model_path: str
    lda_options: MetaDict
    acc_paths: Dict[str, str]


class CalcLdaMlltArguments(MfaArguments):
    """Arguments for :func:`~montreal_forced_aligner.acoustic_modeling.lda.CalcLdaMlltFunction`"""

    dictionaries: List[str]
    feature_strings: Dict[str, str]
    ali_paths: Dict[str, str]
    model_path: str
    lda_options: MetaDict
    macc_paths: Dict[str, str]


class LdaAccStatsFunction(KaldiFunction):
    """
    Multiprocessing function to accumulate LDA stats

    See Also
    --------
    :meth:`.LdaTrainer.lda_acc_stats`
        Main function that calls this function in parallel
    :meth:`.LdaTrainer.lda_acc_stats_arguments`
        Job method for generating arguments for this function
    :kaldi_src:`ali-to-post`
        Relevant Kaldi binary
    :kaldi_src:`weight-silence-post`
        Relevant Kaldi binary
    :kaldi_src:`acc-lda`
        Relevant Kaldi binary

    Parameters
    ----------
    args: :class:`~montreal_forced_aligner.acoustic_modeling.lda.LdaAccStatsArguments`
        Arguments for the function
    """

    progress_pattern = re.compile(r"^LOG.*Done (?P<done>\d+) files, failed for (?P<failed>\d+)$")

    def __init__(self, args: LdaAccStatsArguments):
        super().__init__(args)
        self.dictionaries = args.dictionaries
        self.feature_strings = args.feature_strings
        self.ali_paths = args.ali_paths
        self.model_path = args.model_path
        self.acc_paths = args.acc_paths
        self.lda_options = args.lda_options

    def run(self):
        """Run the function"""
        with open(self.log_path, "w", encoding="utf8") as log_file:
            for dict_name in self.dictionaries:
                ali_path = self.ali_paths[dict_name]
                feature_string = self.feature_strings[dict_name]
                acc_path = self.acc_paths[dict_name]
                ali_to_post_proc = subprocess.Popen(
                    [thirdparty_binary("ali-to-post"), f"ark:{ali_path}", "ark:-"],
                    stderr=log_file,
                    stdout=subprocess.PIPE,
                    env=os.environ,
                )
                weight_silence_post_proc = subprocess.Popen(
                    [
                        thirdparty_binary("weight-silence-post"),
                        "0.0",
                        self.lda_options["silence_csl"],
                        self.model_path,
                        "ark:-",
                        "ark:-",
                    ],
                    stdin=ali_to_post_proc.stdout,
                    stderr=log_file,
                    stdout=subprocess.PIPE,
                    env=os.environ,
                )
                acc_lda_post_proc = subprocess.Popen(
                    [
                        thirdparty_binary("acc-lda"),
                        f"--rand-prune={self.lda_options['random_prune']}",
                        self.model_path,
                        feature_string,
                        "ark,s,cs:-",
                        acc_path,
                    ],
                    stdin=weight_silence_post_proc.stdout,
                    stderr=subprocess.PIPE,
                    encoding="utf8",
                    env=os.environ,
                )
                for line in acc_lda_post_proc.stderr:
                    log_file.write(line)
                    m = self.progress_pattern.match(line.strip())
                    if m:
                        yield int(m.group("done")), int(m.group("failed"))
                self.check_call(acc_lda_post_proc)


class CalcLdaMlltFunction(KaldiFunction):
    """
    Multiprocessing function for estimating LDA with MLLT.

    See Also
    --------
    :meth:`.LdaTrainer.calc_lda_mllt`
        Main function that calls this function in parallel
    :meth:`.LdaTrainer.calc_lda_mllt_arguments`
        Job method for generating arguments for this function
    :kaldi_src:`ali-to-post`
        Relevant Kaldi binary
    :kaldi_src:`weight-silence-post`
        Relevant Kaldi binary
    :kaldi_src:`gmm-acc-mllt`
        Relevant Kaldi binary

    Parameters
    ----------
    args: :class:`~montreal_forced_aligner.acoustic_modeling.lda.CalcLdaMlltArguments`
        Arguments for the function
    """

    progress_pattern = re.compile(r"^LOG.*Average like for this file.*$")

    def __init__(self, args: CalcLdaMlltArguments):
        super().__init__(args)
        self.dictionaries = args.dictionaries
        self.feature_strings = args.feature_strings
        self.ali_paths = args.ali_paths
        self.model_path = args.model_path
        self.macc_paths = args.macc_paths
        self.lda_options = args.lda_options

    def run(self):
        """Run the function"""
        # Estimating MLLT
        with open(self.log_path, "w", encoding="utf8") as log_file:
            for dict_name in self.dictionaries:
                ali_path = self.ali_paths[dict_name]
                feature_string = self.feature_strings[dict_name]
                macc_path = self.macc_paths[dict_name]
                post_proc = subprocess.Popen(
                    [thirdparty_binary("ali-to-post"), f"ark:{ali_path}", "ark:-"],
                    stdout=subprocess.PIPE,
                    stderr=log_file,
                    env=os.environ,
                )

                weight_proc = subprocess.Popen(
                    [
                        thirdparty_binary("weight-silence-post"),
                        "0.0",
                        self.lda_options["silence_csl"],
                        self.model_path,
                        "ark:-",
                        "ark:-",
                    ],
                    stdin=post_proc.stdout,
                    stdout=subprocess.PIPE,
                    stderr=log_file,
                    env=os.environ,
                )
                acc_proc = subprocess.Popen(
                    [
                        thirdparty_binary("gmm-acc-mllt"),
                        f"--rand-prune={self.lda_options['random_prune']}",
                        self.model_path,
                        feature_string,
                        "ark,s,cs:-",
                        macc_path,
                    ],
                    stdin=weight_proc.stdout,
                    stderr=subprocess.PIPE,
                    encoding="utf8",
                    env=os.environ,
                )
                for line in acc_proc.stderr:
                    log_file.write(line)
                    m = self.progress_pattern.match(line.strip())
                    if m:
                        yield 1
                self.check_call(acc_proc)


class LdaTrainer(TriphoneTrainer):
    """
    Triphone trainer

    Parameters
    ----------
    subset : int
        Number of utterances to use, defaults to 10000
    num_leaves : int
        Number of states in the decision tree, defaults to 2500
    max_gaussians : int
        Number of gaussians in the decision tree, defaults to 15000
    lda_dimension : int
        Dimensionality of the LDA matrix
    uses_splices : bool
        Flag to use spliced and LDA calculation
    splice_left_context : int or None
        Number of frames to splice on the left for calculating LDA
    splice_right_context : int or None
        Number of frames to splice on the right for calculating LDA
    random_prune : float
        This is approximately the ratio by which we will speed up the
        LDA and MLLT calculations via randomized pruning

    See Also
    --------
    :class:`~montreal_forced_aligner.acoustic_modeling.triphone.TriphoneTrainer`
        For acoustic model training parsing parameters

    Attributes
    ----------
    mllt_iterations : list
        List of iterations to perform MLLT estimation
    """

    def __init__(
        self,
        subset: int = 10000,
        num_leaves: int = 2500,
        max_gaussians=15000,
        lda_dimension: int = 40,
        uses_splices: bool = True,
        splice_left_context: int = 3,
        splice_right_context: int = 3,
        random_prune=4.0,
        boost_silence: float = 1.0,
        power: float = 0.25,
        **kwargs,
    ):
        super().__init__(
            boost_silence=boost_silence,
            power=power,
            subset=subset,
            num_leaves=num_leaves,
            max_gaussians=max_gaussians,
            **kwargs,
        )
        self.lda_dimension = lda_dimension
        self.random_prune = random_prune
        self.uses_splices = uses_splices
        self.splice_left_context = splice_left_context
        self.splice_right_context = splice_right_context

    def lda_acc_stats_arguments(self) -> List[LdaAccStatsArguments]:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.acoustic_modeling.lda.LdaAccStatsFunction`

        Returns
        -------
        list[:class:`~montreal_forced_aligner.acoustic_modeling.lda.LdaAccStatsArguments`]
            Arguments for processing
        """
        feat_strings = self.worker.construct_feature_proc_strings()
        return [
            LdaAccStatsArguments(
                j.name,
                getattr(self, "db_path", ""),
                os.path.join(self.working_log_directory, f"lda_acc_stats.{j.name}.log"),
                j.dictionary_names,
                feat_strings[j.name],
                j.construct_path_dictionary(self.previous_aligner.working_directory, "ali", "ark"),
                self.previous_aligner.alignment_model_path,
                self.lda_options,
                j.construct_path_dictionary(self.working_directory, "lda", "acc"),
            )
            for j in self.jobs
        ]

    def calc_lda_mllt_arguments(self) -> List[CalcLdaMlltArguments]:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.acoustic_modeling.lda.CalcLdaMlltFunction`

        Returns
        -------
        list[:class:`~montreal_forced_aligner.acoustic_modeling.lda.CalcLdaMlltArguments`]
            Arguments for processing
        """
        feat_strings = self.worker.construct_feature_proc_strings()
        return [
            CalcLdaMlltArguments(
                j.name,
                getattr(self, "db_path", ""),
                os.path.join(
                    self.working_log_directory, f"lda_mllt.{self.iteration}.{j.name}.log"
                ),
                j.dictionary_names,
                feat_strings[j.name],
                j.construct_path_dictionary(self.working_directory, "ali", "ark"),
                self.model_path,
                self.lda_options,
                j.construct_path_dictionary(self.working_directory, "lda", "macc"),
            )
            for j in self.jobs
        ]

    @property
    def train_type(self) -> str:
        """Training identifier"""
        return "lda"

    @property
    def lda_options(self) -> MetaDict:
        """Options for computing LDA"""
        return {
            "lda_dimension": self.lda_dimension,
            "random_prune": self.random_prune,
            "silence_csl": self.silence_csl,
        }

    def compute_calculated_properties(self) -> None:
        """Generate realignment iterations, MLLT estimation iterations, and initial gaussians based on configuration"""
        super().compute_calculated_properties()
        self.mllt_iterations = [2, 4, 6, 12]

    def lda_acc_stats(self) -> None:
        """
        Multiprocessing function that accumulates LDA statistics.

        See Also
        --------
        :func:`~montreal_forced_aligner.acoustic_modeling.lda.LdaAccStatsFunction`
            Multiprocessing helper function for each job
        :meth:`.LdaTrainer.lda_acc_stats_arguments`
            Job method for generating arguments for the helper function
        :kaldi_src:`est-lda`
            Relevant Kaldi binary
        :kaldi_steps:`train_lda_mllt`
            Reference Kaldi script

        """
        worker_lda_path = os.path.join(self.worker.working_directory, "lda.mat")
        lda_path = os.path.join(self.working_directory, "lda.mat")
        if os.path.exists(worker_lda_path):
            os.remove(worker_lda_path)
        arguments = self.lda_acc_stats_arguments()
        with tqdm.tqdm(
            total=self.num_current_utterances, disable=getattr(self, "quiet", False)
        ) as pbar:
            if self.use_mp:
                manager = mp.Manager()
                error_dict = manager.dict()
                return_queue = manager.Queue()
                stopped = Stopped()
                procs = []
                for i, args in enumerate(arguments):
                    function = LdaAccStatsFunction(args)
                    p = KaldiProcessWorker(i, return_queue, function, error_dict, stopped)
                    procs.append(p)
                    p.start()
                while True:
                    try:
                        done, errors = return_queue.get(timeout=1)
                        if stopped.stop_check():
                            continue
                    except Empty:
                        for proc in procs:
                            if not proc.finished.stop_check():
                                break
                        else:
                            break
                        continue
                    pbar.update(done + errors)
                for p in procs:
                    p.join()
                if error_dict:
                    for v in error_dict.values():
                        raise v
            else:
                for args in arguments:
                    function = LdaAccStatsFunction(args)
                    for done, errors in function.run():
                        pbar.update(done + errors)

        log_path = os.path.join(self.working_log_directory, "lda_est.log")
        acc_list = []
        for x in arguments:
            acc_list.extend(x.acc_paths.values())
        with open(log_path, "w", encoding="utf8") as log_file:
            est_lda_proc = subprocess.Popen(
                [
                    thirdparty_binary("est-lda"),
                    f"--dim={self.lda_dimension}",
                    lda_path,
                ]
                + acc_list,
                stderr=log_file,
                env=os.environ,
            )
            est_lda_proc.communicate()
        shutil.copyfile(
            lda_path,
            worker_lda_path,
        )

    def _trainer_initialization(self) -> None:
        """Initialize LDA training"""
        self.uses_splices = True
        self.worker.uses_splices = True
        self.lda_acc_stats()
        self.tree_stats()
        self._setup_tree(initial_mix_up=False)

        self.compile_train_graphs()

        self.convert_alignments()
        os.rename(self.model_path, self.next_model_path)

    def calc_lda_mllt(self) -> None:
        """
        Multiprocessing function that calculates LDA+MLLT transformations.

        See Also
        --------
        :func:`~montreal_forced_aligner.acoustic_modeling.lda.CalcLdaMlltFunction`
            Multiprocessing helper function for each job
        :meth:`.LdaTrainer.calc_lda_mllt_arguments`
            Job method for generating arguments for the helper function
        :kaldi_src:`est-mllt`
            Relevant Kaldi binary
        :kaldi_src:`gmm-transform-means`
            Relevant Kaldi binary
        :kaldi_src:`compose-transforms`
            Relevant Kaldi binary
        :kaldi_steps:`train_lda_mllt`
            Reference Kaldi script

        """
        self.log_info("Re-calculating LDA...")
        arguments = self.calc_lda_mllt_arguments()
        with tqdm.tqdm(
            total=self.num_current_utterances, disable=getattr(self, "quiet", False)
        ) as pbar:
            if self.use_mp:
                manager = mp.Manager()
                error_dict = manager.dict()
                return_queue = manager.Queue()
                stopped = Stopped()
                procs = []
                for i, args in enumerate(arguments):
                    function = CalcLdaMlltFunction(args)
                    p = KaldiProcessWorker(i, return_queue, function, error_dict, stopped)
                    procs.append(p)
                    p.start()
                while True:
                    try:
                        _ = return_queue.get(timeout=1)
                        if stopped.stop_check():
                            continue
                    except Empty:
                        for proc in procs:
                            if not proc.finished.stop_check():
                                break
                        else:
                            break
                        continue
                    pbar.update(1)
                for p in procs:
                    p.join()
                if error_dict:
                    for v in error_dict.values():
                        raise v
            else:
                for args in arguments:
                    function = CalcLdaMlltFunction(args)
                    for _ in function.run():
                        pbar.update(1)

        log_path = os.path.join(
            self.working_log_directory, f"transform_means.{self.iteration}.log"
        )
        previous_mat_path = os.path.join(self.working_directory, "lda.mat")
        new_mat_path = os.path.join(self.working_directory, "lda_new.mat")
        composed_path = os.path.join(self.working_directory, "lda_composed.mat")
        with open(log_path, "a", encoding="utf8") as log_file:
            macc_list = []
            for x in arguments:
                macc_list.extend(x.macc_paths.values())
            subprocess.call(
                [thirdparty_binary("est-mllt"), new_mat_path] + macc_list,
                stderr=log_file,
                env=os.environ,
            )
            subprocess.call(
                [
                    thirdparty_binary("gmm-transform-means"),
                    new_mat_path,
                    self.model_path,
                    self.model_path,
                ],
                stderr=log_file,
                env=os.environ,
            )

            if os.path.exists(previous_mat_path):
                subprocess.call(
                    [
                        thirdparty_binary("compose-transforms"),
                        new_mat_path,
                        previous_mat_path,
                        composed_path,
                    ],
                    stderr=log_file,
                    env=os.environ,
                )
                os.remove(previous_mat_path)
                os.rename(composed_path, previous_mat_path)
            else:
                os.rename(new_mat_path, previous_mat_path)

    def train_iteration(self) -> None:
        """
        Run a single LDA training iteration
        """
        if os.path.exists(self.next_model_path):
            return
        if self.iteration in self.realignment_iterations:
            self.align_iteration()
        if self.iteration in self.mllt_iterations:
            self.calc_lda_mllt()

        self.acc_stats()
        parse_logs(self.working_log_directory)
        if self.iteration <= self.final_gaussian_iteration:
            self.increment_gaussians()
        self.iteration += 1



"""
Utility functions
=================

"""
from __future__ import annotations

import logging
import multiprocessing as mp
import os
import shutil
import sys
import traceback
from queue import Empty
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import ansiwrap
from colorama import Fore, Style

from montreal_forced_aligner.abc import KaldiFunction
from montreal_forced_aligner.data import MfaArguments
from montreal_forced_aligner.exceptions import KaldiProcessingError, ThirdpartyError
from montreal_forced_aligner.models import MODEL_TYPES

__all__ = [
    "thirdparty_binary",
    "log_kaldi_errors",
    "guess_model_type",
    "get_mfa_version",
    "parse_logs",
    "CustomFormatter",
    "Counter",
    "Stopped",
    "ProcessWorker",
    "run_mp",
    "run_non_mp",
]


def get_mfa_version() -> str:
    """
    Get the current MFA version

    Returns
    -------
    str
        MFA version
    """
    try:
        from ._version import version as __version__  # noqa
    except ImportError:
        __version__ = "2.0.0"
    return __version__


def check_third_party():
    """
    Checks whether third party software is available on the path

    Raises
    -------
    :class:`~montreal_forced_aligner.exceptions.ThirdpartyError`
    """
    bin_path = shutil.which("sox")
    if bin_path is None:
        raise ThirdpartyError("sox")
    bin_path = shutil.which("fstcompile")
    if bin_path is None:
        raise ThirdpartyError("fstcompile", open_fst=True)
    bin_path = shutil.which("compute-mfcc-feats")
    if bin_path is None:
        raise ThirdpartyError("compute-mfcc-feats")


def thirdparty_binary(binary_name: str) -> str:
    """
    Generate full path to a given binary name

    Notes
    -----
    With the move to conda, this function is deprecated as conda will manage the path much better

    Parameters
    ----------
    binary_name: str
        Executable to run

    Returns
    -------
    str
        Full path to the executable
    """
    bin_path = shutil.which(binary_name)
    if bin_path is None:
        if binary_name in ["fstcompile", "fstarcsort", "fstconvert"] and sys.platform != "win32":
            raise ThirdpartyError(binary_name, open_fst=True)
        else:
            raise ThirdpartyError(binary_name)
    return bin_path


def log_kaldi_errors(error_logs: List[str], logger: logging.Logger) -> None:
    """
    Save details of Kaldi processing errors to a logger

    Parameters
    ----------
    error_logs: list[str]
        Kaldi log files with errors
    logger: :class:`~logging.Logger`
        Logger to output to
    """
    logger.debug(f"There were {len(error_logs)} kaldi processing files that had errors:")
    for path in error_logs:
        logger.debug("")
        logger.debug(path)
        with open(path, "r", encoding="utf8") as f:
            for line in f:
                logger.debug("\t" + line.strip())


def guess_model_type(path: str) -> List[str]:
    """
    Guess a model type given a path

    Parameters
    ----------
    path: str
        Model archive to guess

    Returns
    -------
    list[str]
        Possible model types that use that extension
    """
    ext = os.path.splitext(path)[1]
    if not ext:
        return []
    possible = []
    for m, mc in MODEL_TYPES.items():
        if ext in mc.extensions:
            possible.append(m)
    return possible


class CustomFormatter(logging.Formatter):
    """
    Custom log formatter class for MFA to highlight messages and incorporate terminal options from
    the global configuration
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from .config import load_global_config

        config = load_global_config()
        self.width = config["terminal_width"]
        use_colors = config.get("terminal_colors", True)
        red = ""
        green = ""
        yellow = ""
        blue = ""
        reset = ""
        if use_colors:
            red = Fore.RED
            green = Fore.GREEN
            yellow = Fore.YELLOW
            blue = Fore.CYAN
            reset = Style.RESET_ALL

        self.FORMATS = {
            logging.DEBUG: (f"{blue}DEBUG{reset} - ", "%(message)s"),
            logging.INFO: (f"{green}INFO{reset} - ", "%(message)s"),
            logging.WARNING: (f"{yellow}WARNING{reset} - ", "%(message)s"),
            logging.ERROR: (f"{red}ERROR{reset} - ", "%(message)s"),
            logging.CRITICAL: (f"{red}CRITICAL{reset} - ", "%(message)s"),
        }

    def format(self, record: logging.LogRecord):
        """
        Format a given log message

        Parameters
        ----------
        record: logging.LogRecord
            Log record to format

        Returns
        -------
        str
            Formatted log message
        """
        log_fmt = self.FORMATS.get(record.levelno)
        return ansiwrap.fill(
            record.getMessage(),
            initial_indent=log_fmt[0],
            subsequent_indent=" " * len(log_fmt[0]),
            width=self.width,
        )


def parse_logs(log_directory: str) -> None:
    """
    Parse the output of a Kaldi run for any errors and raise relevant MFA exceptions

    Parameters
    ----------
    log_directory: str
        Log directory to parse

    Raises
    ------
    KaldiProcessingError
        If any log files contained error lines

    """
    error_logs = []
    for name in os.listdir(log_directory):
        log_path = os.path.join(log_directory, name)
        with open(log_path, "r", encoding="utf8") as f:
            for line in f:
                line = line.strip()
                if "error while loading shared libraries: libopenblas.so.0" in line:
                    raise ThirdpartyError("libopenblas.so.0", open_blas=True)
                for libc_version in ["GLIBC_2.27", "GLIBCXX_3.4.20"]:
                    if libc_version in line:
                        raise ThirdpartyError(libc_version, libc=True)
                if "sox FAIL formats" in line:
                    f = line.split(" ")[-1]
                    raise ThirdpartyError(f, sox=True)
                if line.startswith("ERROR") or line.startswith("ASSERTION_FAILED"):
                    error_logs.append(log_path)
                    break
    if error_logs:
        raise KaldiProcessingError(error_logs)


class Counter(object):
    """
    Multiprocessing counter object for keeping track of progress

    Attributes
    ----------
    val: :func:`~multiprocessing.Value`
        Integer to increment
    lock: :class:`~multiprocessing.Lock`
        Lock for process safety
    """

    def __init__(self, init_val: int = 0):
        self.val = mp.Value("i", init_val)
        self.lock = mp.Lock()

    def increment(self) -> None:
        """Increment the counter"""
        with self.lock:
            self.val.value += 1

    def value(self) -> int:
        """Get the current value of the counter"""
        with self.lock:
            return self.val.value


class Stopped(object):
    """
    Multiprocessing class for detecting whether processes should stop processing and exit ASAP

    Attributes
    ----------
    val: :func:`~multiprocessing.Value`
        0 if not stopped, 1 if stopped
    lock: :class:`~multiprocessing.Lock`
        Lock for process safety
    _source: multiprocessing.Value
        1 if it was a Ctrl+C event that stopped it, 0 otherwise
    """

    def __init__(self, initval: Union[bool, int] = False):
        self.val = mp.Value("i", initval)
        self.lock = mp.Lock()
        self._source = mp.Value("i", 0)

    def stop(self) -> None:
        """Signal that work should stop asap"""
        with self.lock:
            self.val.value = True

    def stop_check(self) -> int:
        """Check whether a process should stop"""
        with self.lock:
            return self.val.value

    def set_sigint_source(self) -> None:
        """Set the source as a ctrl+c"""
        with self.lock:
            self._source.value = True

    def source(self) -> int:
        """Get the source value"""
        with self.lock:
            return self._source.value


class ProcessWorker(mp.Process):
    """
    Multiprocessing function work

    Parameters
    ----------
    job_name: int
        Integer number of job
    job_q: :class:`~multiprocessing.Queue`
        Job queue to pull arguments from
    function: Callable
        Multiprocessing function to call on arguments from job_q
    return_dict: dict
        Dictionary for collecting errors
    stopped: :class:`~montreal_forced_aligner.utils.Stopped`
        Stop check
    return_info: dict[int, Any], optional
        Optional dictionary to fill if the function should return information to main thread
    """

    def __init__(
        self,
        job_name: int,
        job_q: mp.Queue,
        function: Callable,
        return_dict: dict,
        stopped: Stopped,
        return_info: Optional[Dict[int, Any]] = None,
    ):
        mp.Process.__init__(self)
        self.job_name = job_name
        self.function = function
        self.job_q = job_q
        self.return_dict = return_dict
        self.return_info = return_info
        self.stopped = stopped

    def run(self) -> None:
        """
        Run through the arguments in the queue apply the function to them
        """
        try:
            arguments = self.job_q.get(timeout=1)
        except Empty:
            return
        self.job_q.task_done()
        try:
            if isinstance(arguments, MfaArguments):
                result = self.function(arguments)
            else:
                result = self.function(*arguments)
            if self.return_info is not None:
                self.return_info[self.job_name] = result
        except Exception:
            self.stopped.stop()
            self.return_dict["error"] = arguments, Exception(
                traceback.format_exception(*sys.exc_info())
            )


class KaldiProcessWorker(mp.Process):
    """
    Multiprocessing function work

    Parameters
    ----------
    job_name: int
        Integer number of job
    return_q: :class:`~multiprocessing.Queue`
        Queue for returning results
    function: KaldiFunction
        Multiprocessing function to call on arguments from job_q
    error_dict: dict
        Dictionary for collecting errors
    stopped: :class:`~montreal_forced_aligner.utils.Stopped`
        Stop check
    """

    def __init__(
        self,
        job_name: int,
        return_q: mp.Queue,
        function: KaldiFunction,
        error_dict: Dict,
        stopped: Stopped,
    ):
        mp.Process.__init__(self)
        self.job_name = job_name
        self.function = function
        self.return_q = return_q
        self.error_dict = error_dict
        self.stopped = stopped
        self.finished = Stopped()

    def run(self) -> None:
        """
        Run through the arguments in the queue apply the function to them
        """
        try:
            for result in self.function.run():
                self.return_q.put(result)
        except Exception:
            self.stopped.stop()
            exc_type, exc_value, exc_traceback = sys.exc_info()
            self.error_dict[self.job_name] = Exception(
                "\n".join(traceback.format_exception(exc_type, exc_value, exc_traceback))
            )
        finally:
            self.finished.stop()


def run_non_mp(
    function: Callable,
    argument_list: List[Union[Tuple[Any, ...], MfaArguments]],
    log_directory: str,
    return_info: bool = False,
) -> Optional[Dict[Any, Any]]:
    """
    Similar to :func:`run_mp`, but no additional processes are used and the jobs are evaluated in sequential order

    Parameters
    ----------
    function: Callable
        Multiprocessing function to evaluate
    argument_list: list
        List of arguments to process
    log_directory: str
        Directory that all log information from the processes goes to
    return_info: dict, optional
        If the function returns information, supply the return dict to populate

    Returns
    -------
    dict, optional
        If the function returns information, returns the dictionary it was supplied with
    """
    if return_info:
        info = {}
        for i, args in enumerate(argument_list):
            if isinstance(args, MfaArguments):
                info[i] = function(args)
            else:
                info[i] = function(*args)
        parse_logs(log_directory)
        return info

    for args in argument_list:
        if isinstance(args, MfaArguments):
            function(args)
        else:
            function(*args)
    parse_logs(log_directory)


def run_mp(
    function: Callable,
    argument_list: List[Union[Tuple[Any, ...], MfaArguments]],
    log_directory: str,
    return_info: bool = False,
) -> Optional[Dict[int, Any]]:
    """
    Apply a function for each job in parallel

    Parameters
    ----------
    function: Callable
        Multiprocessing function to apply
    argument_list: list
        Arguments for each job
    log_directory: str
        Directory that all log information from the processes goes to
    return_info: dict, optional
        If the function returns information, supply the return dict to populate
    """
    from .config import BLAS_THREADS

    os.environ["OPENBLAS_NUM_THREADS"] = f"{BLAS_THREADS}"
    os.environ["MKL_NUM_THREADS"] = f"{BLAS_THREADS}"
    stopped = Stopped()
    manager = mp.Manager()
    job_queue = manager.Queue()
    return_dict = manager.dict()
    info = None
    if return_info:
        info = manager.dict()
    for a in argument_list:
        job_queue.put(a)
    procs = []
    for i in range(len(argument_list)):
        p = ProcessWorker(i, job_queue, function, return_dict, stopped, info)
        procs.append(p)
        p.start()

    for p in procs:
        p.join()
    if "error" in return_dict:
        _, exc = return_dict["error"]
        raise exc

    parse_logs(log_directory)
    if return_info:
        return info



"""Class definitions for aligning with pretrained acoustic models"""
from __future__ import annotations

import csv
import functools
import multiprocessing as mp
import os
import time
from typing import TYPE_CHECKING, Dict, List, Optional

from sqlalchemy.orm import Session, joinedload, load_only, subqueryload

from montreal_forced_aligner.abc import TopLevelMfaWorker
from montreal_forced_aligner.alignment.base import CorpusAligner
from montreal_forced_aligner.corpus.db import (
    Corpus,
    File,
    PhoneInterval,
    Speaker,
    Utterance,
    WordInterval,
)
from montreal_forced_aligner.exceptions import AlignerError, KaldiProcessingError
from montreal_forced_aligner.helper import align_phones, load_configuration, parse_old_features
from montreal_forced_aligner.models import AcousticModel
from montreal_forced_aligner.online.alignment import (
    OnlineAlignmentArguments,
    OnlineAlignmentFunction,
)
from montreal_forced_aligner.utils import log_kaldi_errors

if TYPE_CHECKING:
    from argparse import Namespace

    from montreal_forced_aligner.abc import MetaDict

__all__ = ["PretrainedAligner"]


class PretrainedAligner(CorpusAligner, TopLevelMfaWorker):
    """
    Class for aligning a dataset using a pretrained acoustic model

    Parameters
    ----------
    acoustic_model_path : str
        Path to acoustic model

    See Also
    --------
    :class:`~montreal_forced_aligner.alignment.base.CorpusAligner`
        For dictionary and corpus parsing parameters and alignment parameters
    :class:`~montreal_forced_aligner.abc.TopLevelMfaWorker`
        For top-level parameters
    """

    def __init__(
        self,
        acoustic_model_path: str,
        **kwargs,
    ):
        self.acoustic_model = AcousticModel(acoustic_model_path)
        kw = self.acoustic_model.parameters
        print("KW", kw)
        kw.update(kwargs)
        super().__init__(**kw)

    @property
    def working_directory(self) -> str:
        """Working directory"""
        return self.workflow_directory

    def setup(self) -> None:
        """Setup for alignment"""
        if self.initialized:
            return
        begin = time.time()
        try:
            os.makedirs(self.working_log_directory, exist_ok=True)
            check = self.check_previous_run()
            if check:
                self.log_debug(
                    "There were some differences in the current run compared to the last one. "
                    "This may cause issues, run with --clean, if you hit an error."
                )
            self.load_corpus()
            if self.excluded_pronunciation_count:
                self.log_warning(
                    f"There were {self.excluded_pronunciation_count} pronunciations in the dictionary that "
                    f"were ignored for containing one of {len(self.excluded_phones)} phones not present in the"
                    f"trained acoustic model.  Please run `mfa validate` to get more details."
                )
            self.acoustic_model.validate(self)
            self.acoustic_model.export_model(self.working_directory)
            import logging

            logger = logging.getLogger(self.identifier)
            self.acoustic_model.log_details(logger)

        except Exception as e:
            if isinstance(e, KaldiProcessingError):
                import logging

                logger = logging.getLogger(self.identifier)
                log_kaldi_errors(e.error_logs, logger)
                e.update_log_file(logger)
            raise
        self.initialized = True
        self.log_debug(f"Setup for alignment in {time.time() - begin} seconds")

    @classmethod
    def parse_parameters(
        cls,
        config_path: Optional[str] = None,
        args: Optional[Namespace] = None,
        unknown_args: Optional[List[str]] = None,
    ) -> MetaDict:
        """
        Parse parameters from a config path or command-line arguments

        Parameters
        ----------
        config_path: str
            Config path
        args: :class:`~argparse.Namespace`
            Command-line arguments from argparse
        unknown_args: list[str], optional
            Extra command-line arguments

        Returns
        -------
        dict[str, Any]
            Configuration parameters
        """
        global_params = {}
        if config_path and os.path.exists(config_path):
            data = load_configuration(config_path)
            data = parse_old_features(data)
            for k, v in data.items():
                if k == "features":
                    global_params.update(v)
                else:
                    if v is None and k in cls.nullable_fields:
                        v = []
                    global_params[k] = v
        global_params.update(cls.parse_args(args, unknown_args))
        return global_params

    @property
    def configuration(self) -> MetaDict:
        """Configuration for aligner"""
        config = super().configuration
        config.update(
            {
                "acoustic_model": self.acoustic_model.name,
            }
        )
        return config

    @property
    def backup_output_directory(self) -> Optional[str]:
        """Backup output directory if overwriting is not allowed"""
        if self.overwrite:
            return None
        return os.path.join(self.working_directory, "textgrids")

    @property
    def workflow_identifier(self) -> str:
        """Aligner identifier"""
        return "pretrained_aligner"

    def evaluate(
        self,
        mapping: Optional[Dict[str, str]] = None,
        output_directory: Optional[str] = None,
    ) -> None:
        """
        Evaluate alignments against a reference directory

        Parameters
        ----------
        reference_directory: str
            Directory containing reference TextGrid alignments
        mapping: dict[str, Union[str, list[str]]], optional
            Mapping between phones that should be considered equal across different phone set types
        output_directory: str, optional
            Directory to save results, if not specified, it will be saved in the log directory
        """
        if output_directory:
            csv_path = os.path.join(output_directory, "alignment_evaluation.csv")
        else:
            csv_path = os.path.join(self.working_log_directory, "alignment_evaluation.csv")
        csv_header = [
            "file",
            "begin",
            "end",
            "speaker",
            "duration",
            "word_count",
            "oov_count",
            "reference_phone_count",
            "score",
            "phone_error_rate",
            "alignment_log_likelihood",
        ]
        if self.alignment_evaluation_done:
            self.log_info("Exporting saved evaluation...")
            with Session(self.db_engine) as session, open(
                csv_path, "w", encoding="utf8", newline=""
            ) as f:
                writer = csv.writer(f)
                writer.writerow(csv_header)
                utterances = session.query(Utterance).options(
                    joinedload(Utterance.speaker, innerjoin=True).load_only(Speaker.name),
                    subqueryload(Utterance.reference_phone_intervals),
                    joinedload(Utterance.file, innerjoin=True).load_only(File.name),
                    load_only(
                        Utterance.id,
                        Utterance.normalized_text,
                        Utterance.oovs,
                        Utterance.alignment_log_likelihood,
                        Utterance.duration,
                        Utterance.begin,
                        Utterance.end,
                        Utterance.alignment_score,
                        Utterance.phone_error_rate,
                    ),
                )
                for u in utterances:
                    word_count = len(u.normalized_text.split())
                    oov_count = len(u.oovs.split())
                    writer.writerow(
                        [
                            u.file.name,
                            u.begin,
                            u.end,
                            u.speaker.name,
                            u.duration,
                            word_count,
                            oov_count,
                            len(u.reference_phone_intervals),
                            u.alignment_score,
                            u.phone_error_rate,
                            u.alignment_log_likelihood,
                        ]
                    )
            return
        # Set up
        self.log_info("Evaluating alignments...")
        self.log_debug(f"Mapping: {mapping}")

        score_count = 0
        score_sum = 0
        phone_edit_sum = 0
        phone_length_sum = 0
        update_mappings = []
        indices = []
        to_comp = []
        score_func = functools.partial(
            align_phones,
            silence_phone=self.optional_silence_phone,
            ignored_phones={self.oov_phone},
            custom_mapping=mapping,
        )
        with Session(self.db_engine) as session:
            unaligned_utts = []
            utterances = session.query(Utterance).options(
                joinedload(Utterance.speaker, innerjoin=True).load_only(Speaker.name),
                subqueryload(Utterance.reference_phone_intervals),
                subqueryload(Utterance.phone_intervals),
                joinedload(Utterance.file, innerjoin=True).load_only(File.name),
                load_only(
                    Utterance.id,
                    Utterance.normalized_text,
                    Utterance.oovs,
                    Utterance.alignment_log_likelihood,
                    Utterance.duration,
                    Utterance.begin,
                    Utterance.end,
                ),
            )
            for u in utterances:
                reference_phone_count = len(u.reference_phone_intervals)
                if not reference_phone_count:
                    continue
                if u.alignment_log_likelihood is None:  # couldn't be aligned
                    phone_error_rate = len(u.reference_phone_intervals)
                    unaligned_utts.append(u)
                    update_mappings.append(
                        {"id": u.id, "alignment_score": None, "phone_error_rate": phone_error_rate}
                    )
                    continue
                reference_phone_labels = [x.as_ctm() for x in u.reference_phone_intervals]
                phone_labels = [x.as_ctm() for x in u.phone_intervals]
                indices.append(u)
                to_comp.append((reference_phone_labels, phone_labels))

            with mp.Pool(self.num_jobs) as pool, open(
                csv_path, "w", encoding="utf8", newline=""
            ) as f:
                writer = csv.writer(f)
                writer.writerow(csv_header)
                for u in unaligned_utts:
                    word_count = len(u.normalized_text.split())
                    oov_count = len(u.oovs.split())
                    writer.writerow(
                        [
                            u.file.name,
                            u.begin,
                            u.end,
                            u.speaker.name,
                            u.duration,
                            word_count,
                            oov_count,
                            reference_phone_count,
                            None,
                            phone_error_rate,
                            None,
                        ]
                    )
                gen = pool.starmap(score_func, to_comp)
                for i, (score, phone_error_rate) in enumerate(gen):
                    if score is None:
                        continue
                    u = indices[i]
                    word_count = len(u.normalized_text.split())
                    oov_count = len(u.oovs.split())
                    reference_phone_count = len(u.reference_phone_intervals)
                    update_mappings.append(
                        {
                            "id": u.id,
                            "alignment_score": score,
                            "phone_error_rate": phone_error_rate,
                        }
                    )
                    writer.writerow(
                        [
                            u.file.name,
                            u.begin,
                            u.end,
                            u.speaker.name,
                            u.duration,
                            word_count,
                            oov_count,
                            reference_phone_count,
                            score,
                            phone_error_rate,
                            u.alignment_log_likelihood,
                        ]
                    )
                    score_count += 1
                    score_sum += score
                    phone_edit_sum += int(phone_error_rate * reference_phone_count)
                    phone_length_sum += reference_phone_count
            session.bulk_update_mappings(Utterance, update_mappings)
            session.query(Corpus).update({"alignment_evaluation_done": True})
            session.commit()
        self.log_info(f"Average overlap score: {score_sum/score_count}")
        self.log_info(f"Average phone error rate: {phone_edit_sum/phone_length_sum}")

    def align_one_utterance(self, utterance: Utterance, session: Session):
        dictionary = utterance.speaker.dictionary
        self.acoustic_model.export_model(self.working_directory)
        pronunciation_dictionary = self.dictionary_mapping[dictionary.name]
        sox_string = utterance.file.sound_file.sox_string
        if not sox_string:
            sox_string = utterance.file.sound_file.sound_file_path
        text_int_path = os.path.join(self.working_directory, "text.int")
        with open(text_int_path, "w", encoding="utf8") as f:
            f.write(f"{utterance.kaldi_id} {utterance.normalized_text_int}\n")
        if utterance.features:
            feats_path = os.path.join(self.working_directory, "feats.scp")
            with open(feats_path, "w", encoding="utf8") as f:
                f.write(f"{utterance.kaldi_id} {utterance.features}\n")
        else:
            wav_path = os.path.join(self.working_directory, "wav.scp")
            segment_path = os.path.join(self.working_directory, "segments.scp")
            with open(wav_path, "w", encoding="utf8") as f:
                f.write(f"{utterance.file_id} {sox_string}\n")
            with open(segment_path, "w", encoding="utf8") as f:
                f.write(
                    f"{utterance.kaldi_id} {utterance.file_id} {utterance.begin} {utterance.end} {utterance.channel}\n"
                )
        if utterance.speaker.cmvn:
            cmvn_path = os.path.join(self.working_directory, "cmvn.scp")
            with open(cmvn_path, "w", encoding="utf8") as f:
                f.write(f"{utterance.speaker.id} {utterance.speaker.cmvn}\n")
        spk2utt_path = os.path.join(self.working_directory, "spk2utt.scp")
        utt2spk_path = os.path.join(self.working_directory, "utt2spk.scp")
        with open(spk2utt_path, "w") as f:
            f.write(f"{utterance.speaker.id} {utterance.kaldi_id}\n")
        with open(utt2spk_path, "w") as f:
            f.write(f"{utterance.kaldi_id} {utterance.speaker.id}\n")

        args = OnlineAlignmentArguments(
            0,
            self.db_path,
            os.path.join(self.working_directory, "align.log"),
            self.working_directory,
            sox_string,
            utterance.to_data(),
            self.mfcc_options,
            self.pitch_options,
            self.feature_options,
            self.align_options,
            self.alignment_model_path,
            self.tree_path,
            pronunciation_dictionary.disambiguation_symbols_int_path,
            dictionary.lexicon_fst_path,
            dictionary.word_boundary_int_path,
            self.reversed_phone_mapping,
            self.optional_silence_phone,
            pronunciation_dictionary.silence_words,
        )
        func = OnlineAlignmentFunction(args)
        word_intervals, phone_intervals, log_likelihood = func.run()
        print(log_likelihood)
        print(word_intervals, phone_intervals)
        session.query(PhoneInterval).filter(PhoneInterval.utterance_id == utterance.id).delete()
        session.query(WordInterval).filter(WordInterval.utterance_id == utterance.id).delete()
        session.flush()
        for wi in word_intervals:
            session.add(WordInterval.from_ctm(wi, utterance))
        for pi in phone_intervals:
            session.add(PhoneInterval.from_ctm(pi, utterance))
        utterance.alignment_log_likelihood = log_likelihood
        session.commit()

    def align(self) -> None:
        """Run the aligner"""
        self.setup()
        done_path = os.path.join(self.working_directory, "done")
        dirty_path = os.path.join(self.working_directory, "dirty")
        if os.path.exists(done_path):
            self.log_info("Alignment already done, skipping.")
            return
        try:
            log_dir = os.path.join(self.working_directory, "log")
            os.makedirs(log_dir, exist_ok=True)
            self.compile_train_graphs()

            self.log_info("Performing first-pass alignment...")
            self.speaker_independent = True
            self.align_utterances()
            self.compile_information()
            if self.uses_speaker_adaptation:
                if self.alignment_model_path.endswith(".mdl"):
                    if os.path.exists(self.alignment_model_path.replace(".mdl", ".alimdl")):
                        raise AlignerError(
                            "Not using speaker independent model when it is available"
                        )
                self.calc_fmllr()

                self.speaker_independent = False
                assert self.alignment_model_path.endswith(".mdl")
                self.log_info("Performing second-pass alignment...")
                self.align_utterances()

                self.compile_information()
        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                import logging

                logger = logging.getLogger(self.identifier)
                log_kaldi_errors(e.error_logs, logger)
                e.update_log_file(logger)
            raise
        with open(done_path, "w"):
            pass


class DictionaryTrainer(PretrainedAligner):
    """
    Aligner for calculating pronunciation probabilities of dictionary entries

    Parameters
    ----------
    calculate_silence_probs: bool
        Flag for whether to calculate silence probabilities, default is False
    min_count: int
        Specifies the minimum count of words to include in derived probabilities,
        affects probabilities of infrequent words more, default is 1

    See Also
    --------
    :class:`~montreal_forced_aligner.alignment.pretrained.PretrainedAligner`
        For dictionary and corpus parsing parameters and alignment parameters
    """

    def __init__(
        self,
        calculate_silence_probs: bool = False,
        min_count: int = 1,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.calculate_silence_probs = calculate_silence_probs
        self.min_count = min_count

    def export_lexicons(
        self, output_directory: str, silence_probabilities: Optional[bool] = False
    ) -> None:
        """
        Generate pronunciation probabilities for the dictionary

        Parameters
        ----------
        output_directory: str
            Directory in which to save new dictionaries
        silence_probabilities: bool
            Flag for whether to save silence probabilities as well

        See Also
        --------
        :func:`~montreal_forced_aligner.alignment.multiprocessing.GeneratePronunciationsFunction`
            Multiprocessing helper function for each job
        :meth:`.CorpusAligner.generate_pronunciations_arguments`
            Job method for generating arguments for helper function

        """
        self.compute_pronunciation_probabilities()
        os.makedirs(output_directory, exist_ok=True)
        for dictionary in self.dictionary_mapping.values():
            dictionary.export_lexicon(
                os.path.join(output_directory, dictionary.name + ".dict"),
                probability=True,
                silence_probabilities=silence_probabilities,
            )



"""Command line functions for training G2P models"""
from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from montreal_forced_aligner.command_line.utils import validate_model_arg
from montreal_forced_aligner.g2p.trainer import PyniniTrainer

if TYPE_CHECKING:
    from argparse import Namespace


__all__ = ["train_g2p", "validate_args", "run_train_g2p"]


def train_g2p(args: Namespace, unknown_args: Optional[List[str]] = None) -> None:
    """
    Run the G2P model training

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Command line arguments
    unknown_args: list[str]
        Optional arguments that will be passed to configuration objects
    """

    trainer = PyniniTrainer(
        dictionary_path=args.dictionary_path,
        temporary_directory=args.temporary_directory,
        **PyniniTrainer.parse_parameters(args.config_path, args, unknown_args)
    )

    try:
        trainer.setup()
        trainer.train()
        trainer.export_model(args.output_model_path)

    except Exception:
        trainer.dirty = True
        raise
    finally:
        trainer.cleanup()


def validate_args(args: Namespace) -> None:
    """
    Validate the command line arguments

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments

    Raises
    ------
    :class:`~montreal_forced_aligner.exceptions.ArgumentError`
        If there is a problem with any arguments
    """
    args.dictionary_path = validate_model_arg(args.dictionary_path, "dictionary")


def run_train_g2p(args: Namespace, unknown: Optional[List[str]] = None) -> None:
    """
    Wrapper function for running G2P model training

    Parameters
    ----------
    args: :class:`~argparse.Namespace`
        Parsed command line arguments
    unknown: list[str]
        Parsed command line arguments to be passed to the configuration objects
    """
    validate_args(args)
    train_g2p(args, unknown)



"""
Validating corpora
==================

"""
from __future__ import annotations

import multiprocessing as mp
import os
import subprocess
import sys
import time
import typing
from decimal import Decimal
from queue import Empty
from typing import TYPE_CHECKING, Dict, List, Optional

import sqlalchemy
import tqdm
from sqlalchemy.orm import Session, joinedload, load_only, subqueryload

from montreal_forced_aligner.acoustic_modeling.trainer import TrainableAligner
from montreal_forced_aligner.alignment import CorpusAligner, PretrainedAligner
from montreal_forced_aligner.alignment.multiprocessing import compile_information_func
from montreal_forced_aligner.corpus.db import (
    Corpus,
    Dictionary,
    File,
    SoundFile,
    Speaker,
    TextFile,
    Utterance,
)
from montreal_forced_aligner.data import MfaArguments
from montreal_forced_aligner.exceptions import ConfigError, KaldiProcessingError
from montreal_forced_aligner.helper import (
    TerminalPrinter,
    comma_join,
    load_configuration,
    load_scp,
)
from montreal_forced_aligner.transcription.transcriber import TranscriberMixin
from montreal_forced_aligner.utils import (
    KaldiFunction,
    KaldiProcessWorker,
    Stopped,
    log_kaldi_errors,
    run_mp,
    run_non_mp,
    thirdparty_binary,
)

if TYPE_CHECKING:
    from argparse import Namespace
    from dataclasses import dataclass

    from .abc import MetaDict
else:
    from dataclassy import dataclass


__all__ = ["TrainingValidator", "PretrainedValidator"]


@dataclass
class TestUtterancesArguments(MfaArguments):
    """Arguments for :class:`~montreal_forced_aligner.validator.TestUtterancesFunction`"""

    feature_strings: Dict[str, str]
    text_int_paths: Dict[str, str]
    model_path: str
    disambiguation_symbols_int_path: str
    score_options: MetaDict
    text_paths: Dict[str, str]
    tree_path: str
    utt2lm_paths: Dict[str, str]
    order: int
    method: str


@dataclass
class TrainSpeakerLmArguments(MfaArguments):
    """Arguments for :class:`~montreal_forced_aligner.validator.TrainSpeakerLmFunction`"""

    word_symbols_paths: Dict[str, str]
    speaker_mapping: Dict[str, List[str]]
    speaker_paths: Dict[str, str]
    oov_word: str
    order: int
    method: str
    target_num_ngrams: int


class TestUtterancesFunction(KaldiFunction):
    """
    Multiprocessing function to test utterance transcriptions with utterance and speaker ngram models

    See Also
    --------
    :kaldi_src:`compile-train-graphs-fsts`
        Relevant Kaldi binary
    :kaldi_src:`gmm-latgen-faster`
        Relevant Kaldi binary
    :kaldi_src:`lattice-oracle`
        Relevant Kaldi binary
    :openfst_src:`farcompilestrings`
        Relevant OpenFst binary
    :ngram_src:`ngramcount`
        Relevant OpenGrm-Ngram binary
    :ngram_src:`ngrammake`
        Relevant OpenGrm-Ngram binary
    :ngram_src:`ngramshrink`
        Relevant OpenGrm-Ngram binary

    Parameters
    ----------
    args: :class:`~montreal_forced_aligner.validator.TestUtterancesArguments`
        Arguments for the function
    """

    def __init__(self, args: TestUtterancesArguments):
        super().__init__(args)
        self.feature_strings = args.feature_strings
        self.text_int_paths = args.text_int_paths
        self.disambiguation_symbols_int_path = args.disambiguation_symbols_int_path
        self.model_path = args.model_path
        self.score_options = args.score_options
        self.text_paths = args.text_paths
        self.tree_path = args.tree_path
        self.utt2lm_paths = args.utt2lm_paths
        self.order = args.order
        self.method = args.method
        self.reversed_word_mapping = {}
        self.word_symbols_paths = {}
        self.lexicon_disambig_fst_paths = {}

    def run(self) -> typing.Generator[typing.Tuple[int, str]]:
        """Run the function"""
        db_engine = sqlalchemy.create_engine(f"sqlite:///{self.db_path}?mode=ro&nolock=1")
        with Session(db_engine) as session:
            for dict_name in self.feature_strings.keys():
                d = (
                    session.query(Dictionary)
                    .options(
                        subqueryload(Dictionary.words),
                        load_only(
                            Dictionary.words_path,
                            Dictionary.lexicon_disambig_fst_path,
                            Dictionary.oov_word,
                        ),
                    )
                    .filter_by(name=dict_name)
                    .first()
                )

                self.oov_word = d.oov_word
                self.word_symbols_paths[dict_name] = d.words_path
                self.lexicon_disambig_fst_paths[dict_name] = d.lexicon_disambig_fst_path

                self.reversed_word_mapping[dict_name] = {}
                for w in d.words:
                    self.reversed_word_mapping[dict_name][w.id] = w.word
        with open(self.log_path, "w") as log_file:
            for dict_name in self.feature_strings.keys():
                feature_string = self.feature_strings[dict_name]
                text_int_path = self.text_int_paths[dict_name]
                disambig_int_path = self.disambiguation_symbols_int_path
                disambig_L_fst_path = self.lexicon_disambig_fst_paths[dict_name]
                utt2lm_path = self.utt2lm_paths[dict_name]
                text_path = self.text_paths[dict_name]
                word_symbols_path = self.word_symbols_paths[dict_name]

                compile_proc = subprocess.Popen(
                    [
                        thirdparty_binary("compile-train-graphs-fsts"),
                        f"--transition-scale={self.score_options['transition_scale']}",
                        f"--self-loop-scale={self.score_options['self_loop_scale']}",
                        f"--read-disambig-syms={disambig_int_path}",
                        "--batch_size=1",
                        self.tree_path,
                        self.model_path,
                        disambig_L_fst_path,
                        "ark:-",
                        "ark:-",
                    ],
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=log_file,
                    env=os.environ,
                )
                latgen_proc = subprocess.Popen(
                    [
                        thirdparty_binary("gmm-latgen-faster"),
                        f"--acoustic-scale={self.score_options['acoustic_scale']}",
                        f"--beam={self.score_options['beam']}",
                        f"--max-active={self.score_options['max_active']}",
                        f"--lattice-beam={self.score_options['lattice_beam']}",
                        f"--word-symbol-table={word_symbols_path}",
                        "--allow-partial",
                        self.model_path,
                        "ark:-",
                        feature_string,
                        "ark:-",
                    ],
                    stderr=log_file,
                    stdin=compile_proc.stdout,
                    stdout=subprocess.PIPE,
                )

                oracle_proc = subprocess.Popen(
                    [
                        thirdparty_binary("lattice-oracle"),
                        "ark:-",
                        f"ark,t:{text_int_path}",
                        "ark,t:-",
                    ],
                    stdin=latgen_proc.stdout,
                    stdout=subprocess.PIPE,
                    env=os.environ,
                    stderr=log_file,
                )
                texts = load_scp(text_path)
                fsts = load_scp(utt2lm_path)
                temp_dir = os.path.dirname(self.log_path)
                for utt, text in texts.items():
                    if not text:
                        continue
                    mod_path = os.path.join(temp_dir, f"{utt}.mod")
                    far_proc = subprocess.Popen(
                        [
                            "farcompilestrings",
                            "--fst_type=compact",
                            f"--unknown_symbol={self.oov_word}",
                            f"--symbols={word_symbols_path}",
                            "--generate_keys=1",
                            "--keep_symbols",
                        ],
                        stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                        stderr=log_file,
                    )
                    count_proc = subprocess.Popen(
                        ["ngramcount", f"--order={self.order}"],
                        stdin=far_proc.stdout,
                        stdout=subprocess.PIPE,
                        stderr=log_file,
                    )
                    with open(mod_path, "wb") as f:
                        make_proc = subprocess.Popen(
                            ["ngrammake", f"--method={self.method}"],
                            stdin=count_proc.stdout,
                            stdout=f,
                            stderr=log_file,
                        )
                    far_proc.stdin.write(" ".join(text).encode("utf8"))
                    far_proc.stdin.flush()
                    far_proc.stdin.close()
                    make_proc.communicate()
                    merge_proc = subprocess.Popen(
                        ["ngrammerge", "--normalize", "--v=10", mod_path, fsts[utt]],
                        stdout=subprocess.PIPE,
                        stderr=log_file,
                    )
                    print_proc = subprocess.Popen(
                        ["fstprint", "--numeric=true", "--v=10"],
                        stderr=log_file,
                        stdin=merge_proc.stdout,
                        stdout=subprocess.PIPE,
                    )
                    # fst = far_proc.stdout.read()
                    fst = print_proc.communicate()[0]
                    compile_proc.stdin.write(f"{utt}\n".encode("utf8"))
                    compile_proc.stdin.write(fst)
                    compile_proc.stdin.write(b"\n")
                    compile_proc.stdin.flush()
                    output = oracle_proc.stdout.readline()
                    output = output.strip().decode("utf8").split(" ")
                    utterance = int(output[0].split("-")[-1])
                    transcript = " ".join(
                        self.reversed_word_mapping[dict_name][int(x)] for x in output[1:]
                    )
                    os.remove(mod_path)
                    yield utterance, transcript

                compile_proc.stdin.close()
                oracle_proc.communicate()
                self.check_call(oracle_proc)


class TrainSpeakerLmFunction(KaldiFunction):
    """
    Multiprocessing function to training small language models for each speaker

    See Also
    --------
    :openfst_src:`farcompilestrings`
        Relevant OpenFst binary
    :ngram_src:`ngramcount`
        Relevant OpenGrm-Ngram binary
    :ngram_src:`ngrammake`
        Relevant OpenGrm-Ngram binary
    :ngram_src:`ngramshrink`
        Relevant OpenGrm-Ngram binary

    Parameters
    ----------
    args: :class:`~montreal_forced_aligner.validator.TrainSpeakerLmArguments`
        Arguments for the function
    """

    def __init__(self, args: TrainSpeakerLmArguments):
        super().__init__(args)
        self.word_symbols_paths = args.word_symbols_paths
        self.speaker_mapping = args.speaker_mapping
        self.speaker_paths = args.speaker_paths
        self.oov_word = args.oov_word
        self.order = args.order
        self.method = args.method
        self.target_num_ngrams = args.target_num_ngrams

    def run(self):
        """Run the function"""
        with open(self.log_path, "w", encoding="utf8") as log_file:

            for dict_name, speakers in self.speaker_mapping.items():
                word_symbols_path = self.word_symbols_paths[dict_name]
                for speaker in speakers:
                    training_path = self.speaker_paths[speaker]
                    base_path = os.path.splitext(training_path)[0]
                    mod_path = base_path + ".mod"
                    far_proc = subprocess.Popen(
                        [
                            "farcompilestrings",
                            "--fst_type=compact",
                            f"--unknown_symbol={self.oov_word}",
                            f"--symbols={word_symbols_path}",
                            "--keep_symbols",
                            training_path,
                        ],
                        stdout=subprocess.PIPE,
                        stderr=log_file,
                    )
                    count_proc = subprocess.Popen(
                        ["ngramcount", f"--order={self.order}"],
                        stdin=far_proc.stdout,
                        stdout=subprocess.PIPE,
                        stderr=log_file,
                    )
                    make_proc = subprocess.Popen(
                        ["ngrammake", "--method=kneser_ney"],
                        stdin=count_proc.stdout,
                        stdout=subprocess.PIPE,
                        stderr=log_file,
                    )
                    shrink_proc = subprocess.Popen(
                        [
                            "ngramshrink",
                            "--method=relative_entropy",
                            f"--target_number_of_ngrams={self.target_num_ngrams}",
                            "--shrink_opt=2",
                            "--theta=0.001",
                            "-",
                            mod_path,
                        ],
                        stdin=make_proc.stdout,
                        stderr=log_file,
                    )
                    shrink_proc.communicate()
                    self.check_call(shrink_proc)
                    assert os.path.exists(mod_path)
                    os.remove(training_path)
                    yield os.path.exists(mod_path)


class ValidationMixin(CorpusAligner, TranscriberMixin):
    """
    Mixin class for performing validation on a corpus

    Parameters
    ----------
    ignore_acoustics: bool
        Flag for whether feature generation and training/alignment should be skipped
    test_transcriptions: bool
        Flag for whether utterance transcriptions should be tested with a unigram language model
    target_num_ngrams: int
        Target number of ngrams from speaker models to use

    See Also
    --------
    :class:`~montreal_forced_aligner.alignment.base.CorpusAligner`
        For corpus, dictionary, and alignment parameters

    Attributes
    ----------
    printer: TerminalPrinter
        Printer for output messages
    """

    def __init__(
        self,
        ignore_acoustics: bool = False,
        test_transcriptions: bool = False,
        target_num_ngrams: int = 100,
        min_word_count: int = 10,
        order: int = 3,
        method: str = "kneser_ney",
        **kwargs,
    ):
        kwargs["clean"] = True
        super().__init__(**kwargs)
        self.ignore_acoustics = ignore_acoustics
        self.test_transcriptions = test_transcriptions
        self.target_num_ngrams = target_num_ngrams
        self.min_word_count = min_word_count
        self.order = order
        self.method = method
        self.printer = TerminalPrinter(print_function=self.log_info)

    def output_utt_fsts(self) -> None:
        """
        Write utterance FSTs
        """

        with Session(self.db_engine) as session:
            for j in self.jobs:
                for dict_name in j.dictionary_names:
                    utterances = (
                        session.query(Utterance)
                        .join(Utterance.speaker)
                        .join(Speaker.dictionary)
                        .options(
                            joinedload(Utterance.speaker).joinedload(Speaker.dictionary),
                            load_only(Utterance.id, Utterance.speaker_id),
                        )
                        .filter(Speaker.job_id == j.name)
                        .filter(Dictionary.name == dict_name)
                        .order_by(Utterance.kaldi_id)
                    )

                    utt2fst_scp_path = os.path.join(
                        self.split_directory, f"utt2lm.{dict_name}.{j.name}.scp"
                    )
                    with open(utt2fst_scp_path, "w", encoding="utf8") as f:
                        for u in utterances:
                            utterance = f"{u.speaker_id}-{u.id}"
                            speaker_lm = os.path.join(
                                self.working_directory, f"{u.speaker_id}.mod"
                            )
                            f.write(f"{utterance} {speaker_lm}\n")

    def train_speaker_lm_arguments(
        self,
    ) -> List[TrainSpeakerLmArguments]:
        """
        Generate Job arguments for :class:`~montreal_forced_aligner.validator.TrainSpeakerLmFunction`

        Returns
        -------
        list[:class:`~montreal_forced_aligner.validator.TrainSpeakerLmArguments`]
            Arguments for processing
        """
        arguments = []
        with Session(self.db_engine) as session:
            for j in self.jobs:
                speaker_mapping = {}
                speaker_paths = {}
                words_symbol_paths = {}

                speakers = (
                    session.query(Speaker)
                    .options(
                        joinedload(Speaker.dictionary, innerjoin=True).load_only(
                            Dictionary.words_path, Dictionary.name
                        )
                    )
                    .filter(Speaker.job_id == j.name)
                )
                for s in speakers:
                    dict_name = s.dictionary.name
                    if dict_name not in speaker_mapping:
                        speaker_mapping[dict_name] = []
                        words_symbol_paths[dict_name] = s.dictionary.words_path
                    speaker_mapping[dict_name].append(s.id)
                    speaker_paths[s.id] = os.path.join(self.working_directory, f"{s.id}.txt")
                arguments.append(
                    TrainSpeakerLmArguments(
                        j.name,
                        getattr(self, "db_path", ""),
                        os.path.join(self.working_log_directory, f"train_lm.{j.name}.log"),
                        words_symbol_paths,
                        speaker_mapping,
                        speaker_paths,
                        self.oov_word,
                        self.order,
                        self.method,
                        self.target_num_ngrams,
                    )
                )
        return arguments

    def test_utterances_arguments(self) -> List[TestUtterancesArguments]:
        """
        Generate Job arguments for :class:`~montreal_forced_aligner.validator.TestUtterancesFunction`

        Returns
        -------
        list[:class:`~montreal_forced_aligner.validator.TestUtterancesArguments`]
            Arguments for processing
        """
        feat_strings = self.construct_feature_proc_strings()

        return [
            TestUtterancesArguments(
                j.name,
                getattr(self, "db_path", ""),
                os.path.join(self.working_log_directory, f"test_utterances.{j.name}.log"),
                feat_strings[j.name],
                j.construct_path_dictionary(self.data_directory, "text", "int.scp"),
                self.model_path,
                self.disambiguation_symbols_int_path,
                self.score_options,
                j.construct_path_dictionary(self.data_directory, "text", "scp"),
                self.tree_path,
                j.construct_path_dictionary(self.data_directory, "utt2lm", "scp"),
                self.order,
                self.method,
            )
            for j in self.jobs
        ]

    @property
    def working_log_directory(self) -> str:
        """Working log directory"""
        return os.path.join(self.working_directory, "log")

    def setup(self):
        """
        Set up the corpus and validator

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        try:
            self.load_corpus()
            self.write_lexicon_information()
            if self.test_transcriptions:
                self.write_lexicon_information(write_disambiguation=True)
            if self.ignore_acoustics:
                self.log_info("Skipping acoustic feature generation")
            else:
                self.generate_features()
            self.calculate_oovs_found()

            if not self.ignore_acoustics and self.test_transcriptions:
                self.initialize_utt_fsts()
            else:
                self.log_info("Skipping transcription testing")
        except Exception as e:
            if isinstance(e, KaldiProcessingError):
                import logging

                logger = logging.getLogger(self.identifier)
                log_kaldi_errors(e.error_logs, logger)
                e.update_log_file(logger)
            raise

    def analyze_setup(self) -> None:
        """
        Analyzes the setup process and outputs info to the console
        """
        begin = time.time()

        with Session(self.db_engine) as session:
            sound_file_count = session.query(SoundFile).count()
            text_file_count = session.query(TextFile).count()
            total_duration = session.query(sqlalchemy.func.sum(Utterance.duration)).scalar()

        total_duration = Decimal(str(total_duration)).quantize(Decimal("0.001"))
        self.log_debug(f"Duration calculation took {time.time() - begin}")

        begin = time.time()
        ignored_count = len(self.no_transcription_files)
        ignored_count += len(self.textgrid_read_errors)
        ignored_count += len(self.decode_error_files)
        self.log_debug(f"Ignored count calculation took {time.time() - begin}")

        self.printer.print_header("Corpus")
        self.printer.print_green_stat(sound_file_count, "sound files")
        self.printer.print_green_stat(text_file_count, "text files")
        if len(self.no_transcription_files):
            self.printer.print_yellow_stat(
                len(self.no_transcription_files),
                "sound files without corresponding transcriptions",
            )
        if len(self.decode_error_files):
            self.printer.print_red_stat(len(self.decode_error_files), "read errors for lab files")
        if len(self.textgrid_read_errors):
            self.printer.print_red_stat(
                len(self.textgrid_read_errors), "read errors for TextGrid files"
            )

        self.printer.print_green_stat(self.num_speakers, "speakers")
        self.printer.print_green_stat(self.num_utterances, "utterances")
        self.printer.print_green_stat(total_duration, "seconds total duration")
        print()
        self.analyze_wav_errors()
        self.analyze_missing_features()
        self.analyze_files_with_no_transcription()
        self.analyze_transcriptions_with_no_wavs()

        if len(self.decode_error_files):
            self.analyze_unreadable_text_files()
        if len(self.textgrid_read_errors):
            self.analyze_textgrid_read_errors()

        self.printer.print_header("Dictionary")
        self.analyze_oovs()

    def analyze_oovs(self) -> None:
        """
        Analyzes OOVs in the corpus and constructs message
        """
        self.printer.print_sub_header("Out of vocabulary words")
        output_dir = self.output_directory
        oov_path = os.path.join(output_dir, "oovs_found.txt")
        utterance_oov_path = os.path.join(output_dir, "utterance_oovs.txt")

        total_instances = 0

        with open(utterance_oov_path, "w", encoding="utf8") as f, Session(
            self.db_engine
        ) as session:
            utterances = (
                session.query(
                    File.name,
                    File.relative_path,
                    Speaker.name,
                    Utterance.begin,
                    Utterance.end,
                    Utterance.oovs,
                )
                .join(Utterance.file)
                .join(Utterance.speaker)
                .filter(Utterance.oovs != None)  # noqa
                .filter(Utterance.oovs != "")
            )

            for file_name, relative_path, speaker_name, begin, end, oovs in utterances:
                total_instances += len(oovs)
                f.write(
                    f"{relative_path +'/' + file_name}, {speaker_name}: {begin}-{end}: {', '.join(oovs)}\n"
                )
                self.oovs_found.update(oovs)
        if self.oovs_found:
            self.save_oovs_found(output_dir)
            self.printer.print_yellow_stat(len(self.oovs_found), "OOV word types")
            self.printer.print_yellow_stat(total_instances, "total OOV tokens")
            lines = [
                "",
                "For a full list of the word types, please see:",
                "",
                self.printer.indent_string + self.printer.colorize(oov_path, "bright"),
                "",
                "For a by-utterance breakdown of missing words, see:",
                "",
                self.printer.indent_string + self.printer.colorize(utterance_oov_path, "bright"),
                "",
            ]
            self.printer.print_info_lines(lines)
        else:
            self.printer.print_info_lines(
                f"There were {self.printer.colorize('no', 'yellow')} missing words from the dictionary. If you plan on using the a model trained "
                "on this dataset to align other datasets in the future, it is recommended that there be at "
                "least some missing words."
            )
        self.printer.print_end_section()

    def analyze_wav_errors(self) -> None:
        """
        Analyzes any sound file issues in the corpus and constructs message
        """
        self.printer.print_sub_header("Sound file read errors")

        output_dir = self.output_directory
        wav_read_errors = self.sound_file_errors
        if wav_read_errors:
            path = os.path.join(output_dir, "sound_file_errors.csv")
            with open(path, "w") as f:
                for p in wav_read_errors:
                    f.write(f"{p}\n")

            self.printer.print_info_lines(
                f"There were {self.printer.colorize(len(wav_read_errors), 'red')} issues reading sound files. "
                f"Please see {self.printer.colorize(path, 'bright')} for a list."
            )
        else:
            self.printer.print_info_lines(
                f"There were {self.printer.colorize('no', 'green')} issues reading sound files."
            )

        self.printer.print_end_section()

    def analyze_missing_features(self) -> None:
        """
        Analyzes issues in feature generation in the corpus and constructs message
        """
        self.printer.print_sub_header("Feature generation")
        if self.ignore_acoustics:
            self.printer.print_info_lines("Acoustic feature generation was skipped.")
            self.printer.print_end_section()
            return
        output_dir = self.output_directory
        with Session(self.db_engine) as session:
            utterances = (
                session.query(File.name, File.relative_path, Utterance.begin, Utterance.end)
                .join(Utterance.file)
                .filter(Utterance.ignored == True)  # noqa
            )
            if utterances.count():
                path = os.path.join(output_dir, "missing_features.csv")
                with open(path, "w") as f:
                    for file_name, relative_path, begin, end in utterances:

                        f.write(f"{relative_path + '/' + file_name},{begin},{end}\n")

                self.printer.print_info_lines(
                    f"There were {self.printer.colorize(len(utterances), 'red')} utterances missing features. "
                    f"Please see {self.printer.colorize(path, 'bright')} for a list."
                )
            else:
                self.printer.print_info_lines(
                    f"There were {self.printer.colorize('no', 'green')} utterances missing features."
                )
            self.printer.print_end_section()

    def analyze_files_with_no_transcription(self) -> None:
        """
        Analyzes issues with sound files that have no transcription files
        in the corpus and constructs message
        """
        self.printer.print_sub_header("Files without transcriptions")
        output_dir = self.output_directory
        if self.no_transcription_files:
            path = os.path.join(output_dir, "missing_transcriptions.csv")
            with open(path, "w") as f:
                for file_path in self.no_transcription_files:
                    f.write(f"{file_path}\n")
            self.printer.print_info_lines(
                f"There were {self.printer.colorize(len(self.no_transcription_files), 'red')} sound files missing transcriptions. "
                f"Please see {self.printer.colorize(path, 'bright')} for a list."
            )
        else:
            self.printer.print_info_lines(
                f"There were {self.printer.colorize('no', 'green')} sound files missing transcriptions."
            )
        self.printer.print_end_section()

    def analyze_transcriptions_with_no_wavs(self) -> None:
        """
        Analyzes issues with transcription that have no sound files
        in the corpus and constructs message
        """
        self.printer.print_sub_header("Transcriptions without sound files")
        output_dir = self.output_directory
        if self.transcriptions_without_wavs:
            path = os.path.join(output_dir, "transcriptions_missing_sound_files.csv")
            with open(path, "w") as f:
                for file_path in self.transcriptions_without_wavs:
                    f.write(f"{file_path}\n")
            self.printer.print_info_lines(
                f"There were {self.printer.colorize(len(self.transcriptions_without_wavs), 'red')} transcription files missing sound files. "
                f"Please see {self.printer.colorize(path, 'bright')} for a list."
            )
        else:
            self.printer.print_info_lines(
                f"There were {self.printer.colorize('no', 'green')} transcription files missing sound files."
            )
        self.printer.print_end_section()

    def analyze_textgrid_read_errors(self) -> None:
        """
        Analyzes issues with reading TextGrid files
        in the corpus and constructs message
        """
        self.printer.print_sub_header("TextGrid read errors")
        output_dir = self.output_directory
        if self.textgrid_read_errors:
            path = os.path.join(output_dir, "textgrid_read_errors.txt")
            with open(path, "w") as f:
                for k, v in self.textgrid_read_errors.items():
                    f.write(
                        f"The TextGrid file {k} gave the following error on load:\n\n{v}\n\n\n"
                    )
            self.printer.print_info_lines(
                [
                    f"There were {self.printer.colorize(len(self.textgrid_read_errors), 'red')} TextGrid files that could not be loaded. "
                    "For details, please see:",
                    "",
                    self.printer.indent_string + self.printer.colorize(path, "bright"),
                ]
            )
        else:
            self.printer.print_info_lines(
                f"There were {self.printer.colorize('no', 'green')} issues reading TextGrids."
            )

        self.printer.print_end_section()

    def analyze_unreadable_text_files(self) -> None:
        """
        Analyzes issues with reading text files
        in the corpus and constructs message
        """
        self.printer.print_sub_header("Text file read errors")
        output_dir = self.output_directory
        if self.decode_error_files:
            path = os.path.join(output_dir, "utf8_read_errors.csv")
            with open(path, "w") as f:
                for file_path in self.decode_error_files:
                    f.write(f"{file_path}\n")
            self.printer.print_info_lines(
                f"There were {self.printer.colorize(len(self.decode_error_files), 'red')} text files that could not be read. "
                f"Please see {self.printer.colorize(path, 'bright')} for a list."
            )
        else:
            self.printer.print_info_lines(
                f"There were {self.printer.colorize('no', 'green')} issues reading text files."
            )

        self.printer.print_end_section()

    def compile_information(self) -> None:
        """
        Compiles information about alignment, namely what the overall log-likelihood was
        and how many files were unaligned.

        See Also
        --------
        :func:`~montreal_forced_aligner.alignment.multiprocessing.compile_information_func`
            Multiprocessing helper function for each job
        :meth:`.AlignMixin.compile_information_arguments`
            Job method for generating arguments for the helper function
        """
        self.log_debug("Analyzing alignment information")
        compile_info_begin = time.time()
        self.collect_alignments()
        jobs = self.compile_information_arguments()

        if self.use_mp:
            alignment_info = run_mp(
                compile_information_func, jobs, self.working_log_directory, True
            )
        else:
            alignment_info = run_non_mp(
                compile_information_func, jobs, self.working_log_directory, True
            )

        avg_like_sum = 0
        avg_like_frames = 0
        average_logdet_sum = 0
        average_logdet_frames = 0
        beam_too_narrow_count = 0
        too_short_count = 0
        unaligned_utts = []
        for data in alignment_info.values():
            unaligned_utts.extend(data["unaligned"])
            beam_too_narrow_count += len(data["unaligned"])
            too_short_count += len(data["too_short"])
            avg_like_frames += data["total_frames"]
            avg_like_sum += data["log_like"] * data["total_frames"]
            if "logdet_frames" in data:
                average_logdet_frames += data["logdet_frames"]
                average_logdet_sum += data["logdet"] * data["logdet_frames"]

        self.printer.print_header("Alignment")
        if not avg_like_frames:
            self.log_debug(
                "No utterances were aligned, this likely indicates serious problems with the aligner."
            )
            self.printer.print_red_stat(0, f"of {self.num_utterances} utterances were aligned")
        else:
            if too_short_count:
                self.printer.print_red_stat(
                    too_short_count, "utterances were too short to be aligned"
                )
            else:
                self.printer.print_green_stat(0, "utterances were too short to be aligned")
            if beam_too_narrow_count:
                self.log_debug(
                    f"There were {beam_too_narrow_count} utterances that could not be aligned with "
                    f"the current beam settings."
                )
                self.printer.print_yellow_stat(
                    beam_too_narrow_count, "utterances that need a larger beam to align"
                )
            else:
                self.printer.print_green_stat(0, "utterances that need a larger beam to align")

            num_utterances = self.num_utterances
            with Session(self.db_engine) as session:
                unaligned_utts = (
                    session.query(Utterance)
                    .options(joinedload(Utterance.file).load_only(File.name))
                    .filter_by(alignment_log_likelihood=None)
                )
                unaligned_count = unaligned_utts.count()
                if unaligned_count:
                    path = os.path.join(self.output_directory, "unalignable_files.csv")
                    with open(path, "w") as f:
                        f.write("file,begin,end,duration,text length\n")
                        for u in unaligned_utts:
                            utt_length_words = u.text.count(" ") + 1
                            f.write(
                                f"{u.file.name},{u.begin},{u.end},{u.duration},{utt_length_words}\n"
                            )
                    self.printer.print_info_lines(
                        [
                            f"There were {self.printer.colorize(unaligned_count, 'red')} unaligned utterances out of {self.printer.colorize(self.num_utterances, 'bright')} after initial training. "
                            f"For details, please see:",
                            "",
                            self.printer.indent_string + self.printer.colorize(path, "bright"),
                        ]
                    )

            self.printer.print_green_stat(
                num_utterances - beam_too_narrow_count - too_short_count,
                "utterances were successfully aligned",
            )
            average_log_like = avg_like_sum / avg_like_frames
            if average_logdet_sum:
                average_log_like += average_logdet_sum / average_logdet_frames
            self.log_debug(f"Average per frame likelihood for alignment: {average_log_like}")
        self.log_debug(f"Compiling information took {time.time() - compile_info_begin}")

    def initialize_utt_fsts(self) -> None:
        """
        Construct utterance FSTs
        """
        if sys.platform != "win32":
            self.log_info("Initializing for testing transcriptions...")
            self.output_utt_fsts()

    @property
    def score_options(self) -> MetaDict:
        return {
            "self_loop_scale": 0.1,
            "transition_scale": 1.0,
            "acoustic_scale": 0.1,
            "beam": 15.0,
            "lattice_beam": 8.0,
            "max_active": 750,
        }

    def train_speaker_lms(self) -> None:
        begin = time.time()
        self.calculate_word_counts()
        log_directory = self.working_log_directory
        os.makedirs(log_directory, exist_ok=True)
        self.log_info("Compiling per speaker biased language models...")
        with Session(self.db_engine) as session:
            speakers = session.query(Speaker).options(
                subqueryload(Speaker.utterances).load_only(Utterance.normalized_text)
            )
            for s in speakers:
                with open(
                    os.path.join(self.working_directory, f"{s.id}.txt"), "w", encoding="utf8"
                ) as f:
                    for u in s.utterances:
                        text = [
                            x if self.word_counts[x] >= self.min_word_count else self.oov_word
                            for x in u.normalized_text.split()
                        ]

                        f.write(" ".join(text) + "\n")
        arguments = self.train_speaker_lm_arguments()
        with tqdm.tqdm(total=self.num_speakers, disable=getattr(self, "quiet", False)) as pbar:
            if self.use_mp:
                manager = mp.Manager()
                error_dict = manager.dict()
                return_queue = manager.Queue()
                stopped = Stopped()
                procs = []

                for i, args in enumerate(arguments):
                    function = TrainSpeakerLmFunction(args)
                    p = KaldiProcessWorker(i, return_queue, function, error_dict, stopped)
                    procs.append(p)
                    p.start()
                while True:
                    try:
                        _ = return_queue.get(timeout=1)
                        if stopped.stop_check():
                            continue
                    except Empty:
                        for proc in procs:
                            if not proc.finished.stop_check():
                                break
                        else:
                            break
                        continue
                    pbar.update(1)
            else:
                self.log_debug("Not using multiprocessing...")
                for args in arguments:
                    function = TrainSpeakerLmFunction(args)
                    for _ in function.run():
                        pbar.update(1)
        self.log_debug(f"Compiling speaker language models took {time.time() - begin}")

    def test_utterance_transcriptions(self) -> None:
        """
        Tests utterance transcriptions with simple unigram models based on the utterance text and frequent
        words in the corpus

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        if sys.platform == "win32":
            self.log_info("Cannot test transcriptions on Windows, please use Linux or Mac.")
            return
        self.log_info("Checking utterance transcriptions...")

        try:
            self.train_speaker_lms()

            self.log_info("Decoding utterances (this will take some time)...")

            begin = time.time()
            log_directory = self.working_log_directory
            os.makedirs(log_directory, exist_ok=True)
            arguments = self.test_utterances_arguments()
            utterance_mapping = []
            with tqdm.tqdm(
                total=self.num_utterances, disable=getattr(self, "quiet", False)
            ) as pbar:
                if self.use_mp:
                    manager = mp.Manager()
                    error_dict = manager.dict()
                    return_queue = manager.Queue()
                    stopped = Stopped()
                    procs = []
                    for i, args in enumerate(arguments):
                        function = TestUtterancesFunction(args)
                        p = KaldiProcessWorker(i, return_queue, function, error_dict, stopped)
                        procs.append(p)
                        p.start()
                    while True:
                        try:
                            utterance, transcript = return_queue.get(timeout=1)
                            if stopped.stop_check():
                                continue
                        except Empty:
                            for proc in procs:
                                if not proc.finished.stop_check():
                                    break
                            else:
                                break
                            continue
                        pbar.update(1)
                        if not utterance or not transcript:
                            continue
                        utterance_mapping.append(
                            {"id": utterance, "transcription_text": transcript}
                        )
                    for p in procs:
                        p.join()
                    if error_dict:
                        for v in error_dict.values():
                            raise v
                else:
                    self.log_debug("Not using multiprocessing...")
                    for args in arguments:
                        function = TestUtterancesFunction(args)
                        for utterance, transcript in function.run():
                            if not utterance or not transcript:
                                continue
                            utterance_mapping.append(
                                {"id": utterance, "transcription_text": transcript}
                            )
                            pbar.update(1)
            with Session(self.db_engine) as session:
                session.bulk_update_mappings(Utterance, utterance_mapping)
            self.log_debug(f"Decoding utterances took {time.time() - begin}")
            self.log_info("Finished decoding utterances!")

            self.printer.print_header("Test transcriptions")
            ser, wer, cer = self.compute_wer()
            if ser < 0.3:
                self.printer.print_green_stat(f"{ser*100:.2f}%", "sentence error rate")
            elif ser < 0.8:
                self.printer.print_yellow_stat(f"{ser*100:.2f}%", "sentence error rate")
            else:
                self.printer.print_red_stat(f"{ser*100:.2f}%", "sentence error rate")

            if wer < 0.25:
                self.printer.print_green_stat(f"{wer*100:.2f}%", "word error rate")
            elif wer < 0.75:
                self.printer.print_yellow_stat(f"{wer*100:.2f}%", "word error rate")
            else:
                self.printer.print_red_stat(f"{wer*100:.2f}%", "word error rate")

            if cer < 0.25:
                self.printer.print_green_stat(f"{cer*100:.2f}%", "character error rate")
            elif cer < 0.75:
                self.printer.print_yellow_stat(f"{cer*100:.2f}%", "character error rate")
            else:
                self.printer.print_red_stat(f"{cer*100:.2f}%", "character error rate")

            self.save_transcription_evaluation(self.output_directory)
            out_path = os.path.join(self.output_directory, "transcription_evaluation.csv")
            print(f"See {self.printer.colorize(out_path, 'bright')} for more details.")

        except Exception as e:
            if isinstance(e, KaldiProcessingError):
                import logging

                logger = logging.getLogger(self.identifier)
                log_kaldi_errors(e.error_logs, logger)
                e.update_log_file(logger)
            raise


class TrainingValidator(TrainableAligner, ValidationMixin):
    """
    Validator class for checking whether a corpus and a dictionary will work together
    for training

    See Also
    --------
    :class:`~montreal_forced_aligner.acoustic_modeling.trainer.TrainableAligner`
        For training configuration
    :class:`~montreal_forced_aligner.validator.ValidationMixin`
        For validation parameters

    Attributes
    ----------
    training_configs: dict[str, :class:`~montreal_forced_aligner.acoustic_modeling.monophone.MonophoneTrainer`]
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.training_configs = {}
        self.add_config("monophone", {})

    @property
    def workflow_identifier(self) -> str:
        """Identifier for validation"""
        return "validate_training"

    @classmethod
    def parse_parameters(
        cls,
        config_path: Optional[str] = None,
        args: Optional[Namespace] = None,
        unknown_args: Optional[List[str]] = None,
    ) -> MetaDict:

        """
        Parse parameters for validation from a config path or command-line arguments

        Parameters
        ----------
        config_path: str
            Config path
        args: :class:`~argparse.Namespace`
            Command-line arguments from argparse
        unknown_args: list[str], optional
            Extra command-line arguments

        Returns
        -------
        dict[str, Any]
            Configuration parameters
        """
        global_params = {}
        training_params = []
        use_default = True
        if config_path:
            data = load_configuration(config_path)
            for k, v in data.items():
                if k == "training":
                    for t in v:
                        for k2, v2 in t.items():
                            if "features" in v2:
                                global_params.update(v2["features"])
                                del v2["features"]
                            training_params.append((k2, v2))
                elif k == "features":
                    if "type" in v:
                        v["feature_type"] = v["type"]
                        del v["type"]
                    global_params.update(v)
                else:
                    if v is None and k in cls.nullable_fields:
                        v = []
                    global_params[k] = v
                if training_params:
                    use_default = False
        if use_default:  # default training configuration
            training_params.append(("monophone", {}))
        if training_params:
            if training_params[0][0] != "monophone":
                raise ConfigError("The first round of training must be monophone.")
        global_params["training_configuration"] = training_params
        global_params.update(cls.parse_args(args, unknown_args))
        return global_params

    def setup(self):
        """
        Set up the corpus and validator

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        if self.initialized:
            return
        try:
            all_begin = time.time()
            self.dictionary_setup()
            self.log_debug(f"Loaded dictionary in {time.time() - all_begin}")

            begin = time.time()
            self._load_corpus()
            self.log_debug(f"Loaded corpus in {time.time() - begin}")

            begin = time.time()
            self.set_lexicon_word_set(self.corpus_word_set)
            self.log_debug(f"Set up lexicon word set in {time.time() - begin}")

            self.calculate_oovs_found()

            begin = time.time()
            self.write_lexicon_information()
            self.log_debug(f"Wrote lexicon information in {time.time() - begin}")

            if self.ignore_acoustics:
                self.log_info("Skipping acoustic feature generation")
            else:

                begin = time.time()
                self.initialize_jobs()
                self.log_debug(f"Initialized jobs in {time.time() - begin}")

                begin = time.time()
                self.create_corpus_split()
                self.log_debug(f"Created corpus split directory in {time.time() - begin}")
                if self.test_transcriptions:
                    begin = time.time()
                    self.write_lexicon_information(write_disambiguation=True)
                    self.log_debug(f"Wrote lexicon information in {time.time() - begin}")
                begin = time.time()
                self.generate_features()
                self.log_debug(f"Generated features in {time.time() - begin}")
                if self.test_transcriptions:
                    begin = time.time()
                    self.initialize_utt_fsts()
                    self.log_debug(f"Initialized utterance FSTs in {time.time() - begin}")
                begin = time.time()
                self.calculate_oovs_found()
                self.log_debug(f"Calculated OOVs in {time.time() - begin}")

            self.initialized = True
        except Exception as e:
            if isinstance(e, KaldiProcessingError):
                import logging

                logger = logging.getLogger(self.identifier)
                log_kaldi_errors(e.error_logs, logger)
                e.update_log_file(logger)
            raise

    def validate(self):
        """
        Performs validation of the corpus
        """
        begin = time.time()
        self.log_debug(f"Setup took {time.time() - begin}")
        self.setup()
        self.analyze_setup()
        self.log_debug(f"Setup took {time.time() - begin}")
        if self.ignore_acoustics:
            self.printer.print_info_lines("Skipping test alignments.")
            return
        self.printer.print_header("Training")
        self.train()
        if self.test_transcriptions:
            self.test_utterance_transcriptions()


class PretrainedValidator(PretrainedAligner, ValidationMixin):
    """
    Validator class for checking whether a corpus, a dictionary, and
    an acoustic model will work together for alignment

    See Also
    --------
    :class:`~montreal_forced_aligner.alignment.pretrained.PretrainedAligner`
        For alignment configuration
    :class:`~montreal_forced_aligner.validator.ValidationMixin`
        For validation parameters
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def workflow_identifier(self) -> str:
        """Identifier for validation"""
        return "validate_pretrained"

    def setup(self):
        """
        Set up the corpus and validator

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.KaldiProcessingError`
            If there were any errors in running Kaldi binaries
        """
        if self.initialized:
            return
        try:
            self.dictionary_setup()
            self._load_corpus()
            self.set_lexicon_word_set(self.corpus_word_set)

            self.calculate_oovs_found()

            if self.ignore_acoustics:
                self.log_info("Skipping acoustic feature generation")
            else:
                self.write_lexicon_information()
                self.initialize_jobs()
                self.create_corpus_split()
                if self.test_transcriptions:
                    self.write_lexicon_information(write_disambiguation=True)
                self.generate_features()
                if self.test_transcriptions:
                    self.initialize_utt_fsts()
                else:
                    self.log_info("Skipping transcription testing")
            self.acoustic_model.validate(self)
            self.acoustic_model.export_model(self.working_directory)
            import logging

            logger = logging.getLogger(self.identifier)
            self.acoustic_model.log_details(logger)

            self.initialized = True
            self.log_info("Finished initializing!")
        except Exception as e:
            if isinstance(e, KaldiProcessingError):
                import logging

                logger = logging.getLogger(self.identifier)
                log_kaldi_errors(e.error_logs, logger)
                e.update_log_file(logger)
            raise

    def align(self) -> None:
        """
        Validate alignment

        """
        done_path = os.path.join(self.working_directory, "done")
        dirty_path = os.path.join(self.working_directory, "dirty")
        if os.path.exists(done_path):
            self.log_debug("Alignment already done, skipping.")
            return
        try:
            log_dir = os.path.join(self.working_directory, "log")
            os.makedirs(log_dir, exist_ok=True)
            self.compile_train_graphs()

            self.log_debug("Performing first-pass alignment...")
            self.speaker_independent = True
            self.align_utterances()
            if self.uses_speaker_adaptation:
                self.log_debug("Calculating fMLLR for speaker adaptation...")
                self.calc_fmllr()

                self.speaker_independent = False
                self.log_debug("Performing second-pass alignment...")
                self.align_utterances()

        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                import logging

                logger = logging.getLogger(self.identifier)
                log_kaldi_errors(e.error_logs, logger)
                e.update_log_file(logger)
            raise
        with open(done_path, "w"):
            pass

    def validate(self) -> None:
        """
        Performs validation of the corpus
        """
        self.setup()
        self.analyze_setup()
        self.analyze_missing_phones()
        if self.ignore_acoustics:
            self.log_info("Skipping test alignments.")
            return
        self.align()
        self.alignment_done = True
        self.compile_information()
        if self.test_transcriptions:
            self.test_utterance_transcriptions()
            self.transcription_done = True
            with self.session() as session:
                session.query(Corpus).update({"transcription_done": True})
                session.commit()

    def analyze_missing_phones(self) -> None:
        """Analyzes dictionary and acoustic model for phones in the dictionary that don't have acoustic models"""
        self.printer.print_sub_header("Acoustic model compatibility")
        if self.excluded_pronunciation_count:
            self.printer.print_yellow_stat(
                len(self.excluded_phones), "phones not in acoustic model"
            )
            self.printer.print_yellow_stat(
                self.excluded_pronunciation_count, "ignored pronunciations"
            )

            phone_string = [self.printer.colorize(x, "red") for x in sorted(self.excluded_phones)]
            self.printer.print_info_lines(
                [
                    "",
                    "Phones missing acoustic models:",
                    "",
                    self.printer.indent_string + comma_join(phone_string),
                ]
            )
        else:
            self.printer.print_info_lines(
                f"There were {self.printer.colorize('no', 'green')} phones in the dictionary without acoustic models."
            )
        self.printer.print_end_section()



"""Class definitions for training G2P models"""
from __future__ import annotations

import functools
import multiprocessing as mp
import operator
import os
import queue
import random
import re
import shutil
import statistics
import subprocess
import sys
import time
import traceback
from typing import Any, Dict, List, NamedTuple, Optional, Set

import tqdm

from montreal_forced_aligner.abc import MetaDict, MfaWorker, TopLevelMfaWorker, TrainerMixin
from montreal_forced_aligner.data import WordData
from montreal_forced_aligner.dictionary.pronunciation import PronunciationDictionaryMixin
from montreal_forced_aligner.exceptions import PyniniAlignmentError
from montreal_forced_aligner.g2p.generator import PyniniValidator
from montreal_forced_aligner.helper import score_g2p
from montreal_forced_aligner.models import G2PModel
from montreal_forced_aligner.utils import Counter, Stopped

try:
    import pynini
    import pywrapfst
    from pynini import Fst, TokenType
    from pywrapfst import convert

    G2P_DISABLED = False

except ImportError:
    pynini = None
    pywrapfst = None
    TokenType = Optional[str]
    Fst = None

    def convert(x):
        """stub function"""
        pass

    G2P_DISABLED = True


Labels = List[Any]

TOKEN_TYPES = ["byte", "utf8"]
INF = float("inf")
RAND_MAX = 32767

__all__ = ["RandomStartWorker", "PyniniTrainer", "G2PTrainer"]


class RandomStart(NamedTuple):
    """Parameters for random starts"""

    idx: int
    seed: int
    g_path: str
    p_path: str
    c_path: str
    tempdir: str
    train_opts: List[str]


class RandomStartWorker(mp.Process):
    """
    Random start worker
    """

    def __init__(
        self,
        job_name: int,
        job_q: mp.Queue,
        return_dict: dict,
        log_file: str,
        error_dict: dict,
        counter: Counter,
        stopped: Stopped,
    ):
        mp.Process.__init__(self)
        self.job_name = job_name
        self.job_q = job_q
        self.return_dict = return_dict
        self.log_file = log_file
        self.error_dict = error_dict
        self.counter = counter
        self.stopped = stopped
        self.finished_signal = Stopped()

    def run(self) -> None:
        """Run the random start worker"""
        with open(self.log_file, "w", encoding="utf8") as log_file:
            while True:
                try:
                    args = self.job_q.get(timeout=1)
                except queue.Empty:
                    break
                self.job_q.task_done()
                if self.stopped.stop_check():
                    continue
                try:
                    start = time.time()
                    # Randomize channel model.
                    c_path = os.path.join(args.tempdir, f"c-{args.seed:05d}.fst")
                    fst_path = os.path.join(args.tempdir, f"t-{args.seed:05d}.fst")
                    likelihood_path = fst_path.replace(".fst", ".like")
                    if not os.path.exists(fst_path):
                        cmd = [
                            "baumwelchrandomize",
                            f"--seed={args.seed}",
                            args.c_path,
                            c_path,
                        ]
                        subprocess.check_call(cmd, stderr=log_file)
                        random_end = time.time()
                        log_file.write(
                            f"{args.seed} randomization took {random_end - start} seconds\n"
                        )
                        # Train on randomized channel model.

                        likelihood = INF
                        cmd = [
                            "baumwelchtrain",
                            *args.train_opts,
                            args.g_path,
                            args.p_path,
                            c_path,
                            fst_path,
                        ]
                        log_file.write(f"{args.seed} train command: {' '.join(cmd)}\n")
                        log_file.flush()
                        with subprocess.Popen(cmd, stderr=subprocess.PIPE, text=True) as proc:
                            # Parses STDERR to capture the likelihood.
                            for line in proc.stderr:  # type: ignore
                                log_file.write(line)
                                log_file.flush()
                                line = line.rstrip()
                                match = re.match(r"INFO: Iteration \d+: (-?\d*(\.\d*)?)", line)
                                assert match, line
                                likelihood = float(match.group(1))
                                self.counter.increment()
                            with open(likelihood_path, "w") as f:
                                f.write(str(likelihood))
                        log_file.write(
                            f"{args.seed} training took {time.time() - random_end} seconds\n"
                        )
                    else:
                        with open(likelihood_path, "r") as f:
                            likelihood = f.read().strip()
                    self.return_dict[fst_path] = likelihood
                except Exception:
                    self.stopped.stop()
                    self.finished_signal.stop()
                    self.error_dict[args.idx] = args, Exception(
                        traceback.format_exception(*sys.exc_info())
                    )
        self.finished_signal.stop()
        return


class G2PTrainer(MfaWorker, TrainerMixin):
    """
    Abstract mixin class for G2P training

    Parameters
    ----------
    validation_proportion: float
        Proportion of words to use as the validation set, defaults to 0.1, only used if ``evaluate`` is True
    num_pronunciations: int
        Number of pronunciations to generate
    evaluation_mode: bool
        Flag for whether to evaluate the model performance on an validation set

    See Also
    --------
    :class:`~montreal_forced_aligner.abc.MfaWorker`
        For base MFA parameters
    :class:`~montreal_forced_aligner.abc.TrainerMixin`
        For base trainer parameters
    :class:`~montreal_forced_aligner.dictionary.pronunciation.PronunciationDictionaryMixin`
        For pronunciation dictionary parameters

    Attributes
    ----------
    g2p_training_dictionary: dict[str, list[str]]
        Dictionary of words to pronunciations to train from
    g2p_validation_dictionary: dict[str, list[str]]
        Dictionary of words to pronunciations to validate performance against
    g2p_graphemes: set[str]
        Set of graphemes in the training set
    """

    def __init__(
        self,
        validation_proportion: float = 0.1,
        num_pronunciations: int = 0,
        evaluation_mode: bool = False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.evaluation_mode = evaluation_mode
        self.validation_proportion = validation_proportion
        self.num_pronunciations = num_pronunciations
        self.g2p_training_dictionary = {}
        self.g2p_validation_dictionary = None
        self.g2p_training_graphemes = set()
        self.g2p_validation_graphemes = set()
        self.g2p_training_phones = set()
        self.g2p_validation_phones = set()


class PyniniTrainer(PronunciationDictionaryMixin, G2PTrainer, TopLevelMfaWorker):
    """
    Top-level G2P trainer that uses Pynini functionality

    Parameters
    ----------
    order: int
        Order of the ngram model, defaults to 7
    random_starts: int
        Number of random starts to use in initialization, defaults to 25
    seed: int
        Seed for randomization, defaults to 1917
    delta: float
        Comparison/quantization delta for Baum-Welch training, defaults to 1/1024
    alpha: float
        Step size reduction power parameter for Baum-Welch training;
        full standard batch EM is run (not stepwise) if set to 0, defaults to 1.0
    batch_size:int
        Batch size for Baum-Welch training, defaults to 200
    num_iterations:int
        Maximum number of iterations to use in Baum-Welch training, defaults to 10
    smoothing_method:str
        Smoothing method for the ngram model, defaults to "kneser_ney"
    pruning_method:str
        Pruning method for pruning the ngram model, defaults to "relative_entropy"
    model_size: int
        Target number of ngrams for pruning, defaults to 1000000
    insertions: bool
        Flag for whether to allow for insertions, default True
    deletions: bool
        Flag for whether to allow for deletions, default True
    fst_default_cache_gc: str
        String to pass to OpenFst binaries for GC behavior
    fst_default_cache_gc_limit: str
        String to pass to OpenFst binaries for GC behavior

    See Also
    --------
    :class:`~montreal_forced_aligner.g2p.trainer.G2PTrainer`
        For base G2P training parameters
    :class:`~montreal_forced_aligner.abc.TopLevelMfaWorker`
        For top-level parameters
    """

    _compactor = functools.partial(convert, fst_type="compact_string")

    def __init__(
        self,
        order: int = 7,
        random_starts: int = 25,
        seed: int = 1917,
        delta: float = 1 / 1024,
        alpha: float = 1.0,
        batch_size: int = 800,
        num_iterations: int = 10,
        smoothing_method: str = "kneser_ney",
        pruning_method: str = "relative_entropy",
        model_size: int = 1000000,
        insertions: bool = True,
        deletions: bool = True,
        fst_default_cache_gc="",
        fst_default_cache_gc_limit="",
        **kwargs,
    ):
        self._data_source = os.path.splitext(os.path.basename(kwargs["dictionary_path"]))[0]
        super().__init__(**kwargs)
        self.order = order
        self.random_starts = random_starts
        self.seed = seed
        self.delta = delta
        self.alpha = alpha
        self.batch_size = batch_size
        self.num_iterations = num_iterations
        self.smoothing_method = smoothing_method
        self.pruning_method = pruning_method
        self.model_size = model_size
        self.insertions = insertions
        self.deletions = deletions
        self.fst_default_cache_gc = fst_default_cache_gc
        self.fst_default_cache_gc_limit = fst_default_cache_gc_limit
        self.g_path = os.path.join(self.working_directory, "g.far")
        self.p_path = os.path.join(self.working_directory, "p.far")
        self.c_path = os.path.join(self.working_directory, "c.fst")
        self.align_path = os.path.join(self.working_directory, "align.fst")
        self.afst_path = os.path.join(self.working_directory, "afst.far")
        self.wer = None
        self.ler = None

    @property
    def data_directory(self) -> str:
        """Data directory for trainer"""
        return self.working_directory

    @property
    def data_source_identifier(self) -> str:
        """Dictionary name"""
        return self._data_source

    @property
    def workflow_identifier(self) -> str:
        """Identifier for Pynini G2P trainer"""
        return "pynini_train_g2p"

    @property
    def configuration(self) -> MetaDict:
        """Configuration for G2P trainer"""
        config = super().configuration
        config.update({"dictionary_path": self.dictionary_model.path})
        return config

    def train_iteration(self) -> None:
        """Train iteration, not used"""
        pass

    def setup(self) -> None:
        """Setup for G2P training"""
        if self.initialized:
            return
        os.makedirs(self.working_log_directory, exist_ok=True)
        self.g2p_training_dictionary: Dict[str, WordData] = self.actual_words
        self.initialize_training()
        self.initialized = True

    @property
    def architecture(self) -> str:
        """Pynini"""
        return "pynini"

    @property
    def meta(self) -> MetaDict:
        """Metadata for exported G2P model"""
        from datetime import datetime

        from ..utils import get_mfa_version

        m = {
            "version": get_mfa_version(),
            "architecture": self.architecture,
            "train_date": str(datetime.now()),
            "phones": sorted(self.non_silence_phones),
            "graphemes": self.graphemes,
            "evaluation": {},
            "training": {
                "num_words": len(self.g2p_training_dictionary),
                "num_graphemes": len(self.graphemes),
                "num_phones": len(self.non_silence_phones),
            },
        }

        if self.evaluation_mode:
            m["evaluation"]["num_words"] = len(self.g2p_validation_dictionary)
            m["evaluation"]["word_error_rate"] = self.wer
            m["evaluation"]["phone_error_rate"] = self.ler
        return m

    @property
    def input_path(self) -> str:
        """Path to temporary file to store training data"""
        return os.path.join(self.working_directory, "input.txt")

    def initialize_training(self) -> None:
        """Initialize training G2P model"""
        random.seed(self.seed)
        if self.evaluation_mode:
            word_dict = self.g2p_training_dictionary
            words = sorted(word_dict.keys())
            total_items = len(words)
            validation_items = int(total_items * self.validation_proportion)
            validation_words = random.sample(words, validation_items)
            self.g2p_training_dictionary = {
                k: v for k, v in word_dict.items() if k not in validation_words
            }
            self.g2p_validation_dictionary = {
                k: v for k, v in word_dict.items() if k in validation_words
            }
            if self.debug:
                with open(
                    os.path.join(self.working_directory, "validation_set.txt"),
                    "w",
                    encoding="utf8",
                ) as f:
                    for word in self.g2p_validation_dictionary:
                        f.write(word + "\n")

        if not os.path.exists(self.sym_path):
            phones_path = os.path.join(self.working_directory, "phones_only.txt")
            with open(self.input_path, "w", encoding="utf8") as f2, open(
                phones_path, "w", encoding="utf8"
            ) as phonef:
                for word, v in self.g2p_training_dictionary.items():
                    if re.match(r"\W", word) is not None:
                        continue
                    self.g2p_training_graphemes.update(word)
                    for v2 in v:
                        self.g2p_training_phones.update(v2.pronunciation)
                        f2.write(f"{word}\t{' '.join(v2.pronunciation)}\n")
                        phonef.write(f"{' '.join(v2.pronunciation)}\n")
            subprocess.call(["ngramsymbols", phones_path, self.sym_path])
            if not self.debug:
                os.remove(phones_path)
        self.log_debug(f"Graphemes in training data: {sorted(self.g2p_training_graphemes)}")
        self.log_debug(f"Phones in training data: {sorted(self.g2p_training_phones)}")
        if self.evaluation_mode:
            for k, v in self.g2p_validation_dictionary.items():
                self.g2p_validation_graphemes.update(k)
                for v2 in v:
                    self.g2p_validation_phones.update(v2.pronunciation)
            self.log_debug(
                f"Graphemes in validation data: {sorted(self.g2p_validation_graphemes)}"
            )
            self.log_debug(f"Phones in validation data: {sorted(self.g2p_validation_phones)}")
            grapheme_diff = sorted(self.g2p_validation_graphemes - self.g2p_training_graphemes)
            phone_diff = sorted(self.g2p_validation_phones - self.g2p_training_phones)
            if grapheme_diff:
                self.log_warning(
                    f"The following graphemes appear only in the validation set: {', '.join(grapheme_diff)}"
                )
            if phone_diff:
                self.log_warning(
                    f"The following phones appear only in the validation set: {', '.join(phone_diff)}"
                )

    def clean_up(self) -> None:
        """
        Clean up temporary files
        """
        if self.debug:
            return
        for name in os.listdir(self.working_directory):
            path = os.path.join(self.working_directory, name)
            if os.path.isdir(path):
                shutil.rmtree(path, ignore_errors=True)
            elif not name.endswith(".log"):
                os.remove(path)

    def generate_model(self) -> None:
        """
        Generate an ngram G2P model from FAR strings
        """
        assert os.path.exists(self.far_path)
        if os.path.exists(self.fst_path):
            self.log_info("Model building already done, skipping!")
            return
        with open(
            os.path.join(self.working_log_directory, "model.log"), "w", encoding="utf8"
        ) as logf:
            ngram_count_path = os.path.join(self.working_directory, "ngram.count")
            ngram_make_path = os.path.join(self.working_directory, "ngram.make")
            ngram_shrink_path = os.path.join(self.working_directory, "ngram.shrink")
            ngramcount_proc = subprocess.Popen(
                [
                    "ngramcount",
                    "--require_symbols=false",
                    f"--order={self.order}",
                    self.far_path,
                    ngram_count_path,
                ],
                stderr=logf,
            )
            ngramcount_proc.communicate()

            ngrammake_proc = subprocess.Popen(
                [
                    "ngrammake",
                    f"--method={self.smoothing_method}",
                    ngram_count_path,
                    ngram_make_path,
                ],
                stderr=logf,
            )
            ngrammake_proc.communicate()

            ngramshrink_proc = subprocess.Popen(
                [
                    "ngramshrink",
                    f"--method={self.pruning_method}",
                    f"--target_number_of_ngrams={self.model_size}",
                    ngram_make_path,
                    ngram_shrink_path,
                ],
                stderr=logf,
            )
            ngramshrink_proc.communicate()

            fstencode_proc = subprocess.Popen(
                ["fstencode", "--decode", ngram_shrink_path, self.encoder_path, self.fst_path],
                stderr=logf,
            )
            fstencode_proc.communicate()

        os.remove(ngram_count_path)
        os.remove(ngram_make_path)
        os.remove(ngram_shrink_path)

    def export_model(self, output_model_path: str) -> None:
        """
        Export G2P model to specified path

        Parameters
        ----------
        output_model_path:str
            Path to export model
        """
        directory, filename = os.path.split(output_model_path)
        basename, _ = os.path.splitext(filename)
        models_temp_dir = os.path.join(self.working_directory, "model_archive_temp")
        model = G2PModel.empty(basename, root_directory=models_temp_dir)
        model.add_meta_file(self)
        model.add_fst_model(self.working_directory)
        model.add_sym_path(self.working_directory)
        if directory:
            os.makedirs(directory, exist_ok=True)
        basename, _ = os.path.splitext(output_model_path)
        model.dump(basename)
        model.clean_up()
        # self.clean_up()
        self.log_info(f"Saved model to {output_model_path}")

    @property
    def fst_path(self):
        """Internal temporary FST file"""
        return os.path.join(self.working_directory, "model.fst")

    @property
    def far_path(self):
        """Internal temporary FAR file"""
        return os.path.join(self.working_directory, f"{self.data_source_identifier}.far")

    @property
    def encoder_path(self):
        """Internal temporary encoder file"""
        return os.path.join(self.working_directory, f"{self.data_source_identifier}.enc")

    @property
    def sym_path(self):
        """Internal temporary symbol file"""
        return os.path.join(self.working_directory, "phones.sym")

    def align(self):
        """Runs the entire alignment regimen."""
        self._lexicon_covering()
        self._alignments()
        self._encode()

    @staticmethod
    def _label_union(labels: Set[int], epsilon: bool) -> Fst:
        """Creates FSA over a union of the labels."""
        side = pynini.Fst()
        src = side.add_state()
        side.set_start(src)
        dst = side.add_state()
        if epsilon:
            labels.add(0)
        one = pynini.Weight.one(side.weight_type())
        for label in labels:
            side.add_arc(src, pynini.Arc(label, label, one, dst))
        side.set_final(dst)
        assert side.verify(), "FST is ill-formed"
        return side

    @staticmethod
    def _narcs(f: Fst) -> int:
        """Computes the number of arcs in an FST."""
        return sum(f.num_arcs(state) for state in f.states())

    NON_SYMBOL = ("byte", "utf8")

    def _lexicon_covering(
        self,
    ) -> None:
        """Builds covering grammar and lexicon FARs."""
        # Sets of labels for the covering grammar.
        g_labels: Set[int] = set()
        p_labels: Set[int] = set()
        self.log_info("Constructing grapheme and phoneme FARs")
        input_token_type = "utf8"
        output_token_type = pynini.SymbolTable.read_text(self.sym_path)
        g_writer = pywrapfst.FarWriter.create(self.g_path)
        p_writer = pywrapfst.FarWriter.create(self.p_path)
        with open(self.input_path, "r") as source:
            for (linenum, line) in enumerate(source, 1):
                key = f"{linenum:08x}"
                (g, p) = line.rstrip().split("\t", 1)
                # For both G and P, we compile a FSA, store the labels, and
                # then write the compact version to the FAR.
                g_fst = pynini.accep(g, token_type=input_token_type)
                g_labels.update(g_fst.paths().ilabels())
                g_writer[key] = self._compactor(g_fst)
                p_fst = pynini.accep(p, token_type=output_token_type)
                p_labels.update(p_fst.paths().ilabels())
                p_writer[key] = self._compactor(p_fst)
        self.log_info(f"Processed {linenum:,d} examples")
        self.log_info("Constructing covering grammar")
        self.log_info(f"{len(g_labels)} unique graphemes")
        g_side = self._label_union(g_labels, self.insertions)
        self.log_info(f"{len(p_labels)} unique phones")
        p_side = self._label_union(p_labels, self.deletions)
        # The covering grammar is given by (G job_name P)^*.
        covering = pynini.cross(g_side, p_side).closure().optimize()
        assert covering.num_states() == 1, "Covering grammar FST is ill-formed"
        self.log_info(
            f"Covering grammar has {PyniniTrainer._narcs(covering):,d} arcs",
        )
        covering.write(self.c_path)

    def _alignments(self) -> None:
        """Trains the aligner and constructs the alignments FAR."""
        if not os.path.exists(self.align_path):
            self.log_info("Training aligner")
            train_opts = []
            if self.batch_size:
                train_opts.append(f"--batch_size={self.batch_size}")
            if self.delta:
                train_opts.append(f"--delta={self.delta}")
            if self.fst_default_cache_gc:
                train_opts.append(f"--fst_default_cache_gc={self.fst_default_cache_gc}")
            if self.fst_default_cache_gc_limit:
                train_opts.append(
                    f"--fst_default_cache_gc_limit={self.fst_default_cache_gc_limit}"
                )
            if self.alpha:
                train_opts.append(f"--alpha={self.alpha}")
            if self.num_iterations:
                train_opts.append(f"--max_iters={self.num_iterations}")
            # Constructs the actual command vectors (plus an index for logging
            # purposes).
            random.seed(self.seed)
            starts = [
                (
                    RandomStart(
                        idx,
                        seed,
                        self.g_path,
                        self.p_path,
                        self.c_path,
                        self.working_directory,
                        train_opts,
                    )
                )
                for (idx, seed) in enumerate(
                    random.sample(range(1, RAND_MAX), self.random_starts), 1
                )
            ]
            stopped = Stopped()
            num_commands = len(starts)
            job_queue = mp.JoinableQueue()

            # Actually runs starts.
            self.log_info("Calculating alignments...")
            begin = time.time()
            max_value = num_commands * self.num_iterations
            with tqdm.tqdm(
                total=num_commands * self.num_iterations, disable=getattr(self, "quiet", False)
            ) as pbar:
                for start in starts:
                    job_queue.put(start)
                manager = mp.Manager()
                error_dict = manager.dict()
                return_dict = manager.dict()
                procs = []
                counter = Counter()
                for i in range(self.num_jobs):
                    log_path = os.path.join(self.working_log_directory, f"baumwelch.{i}.log")
                    p = RandomStartWorker(
                        i,
                        job_queue,
                        return_dict,
                        log_path,
                        error_dict,
                        counter,
                        stopped,
                    )
                    procs.append(p)
                    p.start()

                value = 0
                last_value = 0
                if len(self.g2p_training_dictionary) > 100000:
                    sleep_increment = 10
                else:
                    sleep_increment = 2
                while value < max_value:
                    time.sleep(sleep_increment)
                    if stopped.stop_check():
                        break
                    for proc in procs:
                        if not proc.finished_signal.stop_check():
                            break
                    else:
                        break
                    value = counter.value()
                    if value != last_value:
                        pbar.update(value - last_value)
                        last_value = value
                job_queue.join()
                for p in procs:
                    p.join()
            if error_dict:
                raise PyniniAlignmentError(error_dict)
            (best_fst, best_likelihood) = min(return_dict.items(), key=operator.itemgetter(1))
            self.log_info(f"Best likelihood: {best_likelihood}")
            self.log_debug(
                f"Ran {self.random_starts} random starts in {time.time() - begin} seconds"
            )
            # Moves best likelihood solution to the requested location.
            shutil.move(best_fst, self.align_path)
        self.log_info("Computing alignments")
        cmd = ["baumwelchdecode"]
        if self.fst_default_cache_gc:
            cmd.append(f"--fst_default_cache_gc={self.fst_default_cache_gc}")
        if self.fst_default_cache_gc_limit:
            cmd.append(f"--fst_default_cache_gc_limit={self.fst_default_cache_gc_limit}")
        cmd.append(self.g_path)
        cmd.append(self.p_path)
        cmd.append(self.align_path)
        cmd.append(self.afst_path)
        self.log_debug(f"Subprocess call: {cmd}")
        subprocess.check_call(cmd)

    def _encode(self) -> None:
        """Encodes the alignments."""
        self.log_info("Encoding the alignments as FSAs")
        encoder = pywrapfst.EncodeMapper(encode_labels=True)
        a_reader = pywrapfst.FarReader.open(self.afst_path)
        a_writer = pywrapfst.FarWriter.create(self.far_path)
        # Curries converter function for the FAR.
        converter = functools.partial(pywrapfst.convert, fst_type="vector")
        while not a_reader.done():
            key = a_reader.get_key()
            fst = converter(a_reader.get_fst())
            fst.encode(encoder)
            a_writer[key] = self._compactor(fst)
            try:
                next(a_reader)
            except StopIteration:
                break
        encoder.write(self.encoder_path)
        self.log_info(f"Success! FAR path: {self.far_path}; encoder path: {self.encoder_path}")

    def train(self) -> None:
        """
        Train a G2P model
        """
        os.makedirs(self.working_log_directory, exist_ok=True)
        begin = time.time()
        if os.path.exists(self.far_path) and os.path.exists(self.encoder_path):
            self.log_info("Alignment already done, skipping!")
        else:
            self.align()
            self.log_debug(
                f"Aligning {len(self.g2p_training_dictionary)} words took {time.time() - begin} seconds"
            )
        begin = time.time()
        self.generate_model()
        self.log_debug(
            f"Generating model for {len(self.g2p_training_dictionary)} words took {time.time() - begin} seconds"
        )
        self.finalize_training()

    def finalize_training(self) -> None:
        """Finalize training"""
        if self.evaluation_mode:
            self.evaluate_g2p_model()

    def evaluate_g2p_model(self) -> None:
        """
        Validate the G2P model against held out data
        """
        temp_model_path = os.path.join(self.working_log_directory, "g2p_model.zip")
        self.export_model(temp_model_path)

        gen = PyniniValidator(
            g2p_model_path=temp_model_path,
            word_list=list(self.g2p_validation_dictionary.keys()),
            temporary_directory=os.path.join(self.working_directory, "validation"),
            num_jobs=self.num_jobs,
            num_pronunciations=self.num_pronunciations,
        )
        output = gen.generate_pronunciations()
        self.compute_validation_errors(output)

    def compute_validation_errors(
        self,
        hypothesis_values: Dict[str, WordData],
    ):
        """
        Computes validation errors

        Parameters
        ----------
        hypothesis_values: dict[str, list[str]]
            Hypothesis labels
        """
        begin = time.time()
        # Word-level measures.
        correct = 0
        incorrect = 0
        # Label-level measures.
        total_edits = 0
        total_length = 0
        # Since the edit distance algorithm is quadratic, let's do this with
        # multiprocessing.
        self.log_debug(f"Processing results for {len(hypothesis_values)} hypotheses")
        to_comp = []
        hyp_pron_count = 0
        gold_pron_count = 0
        for word, gold in self.g2p_validation_dictionary.items():
            if word not in hypothesis_values:
                incorrect += 1
                gold_length = statistics.mean(len(x) for x in gold.pronunciations)
                total_edits += gold_length
                total_length += gold_length
                continue
            hyp = hypothesis_values[word]
            for h in hyp.pronunciations:
                if h in gold.pronunciations:
                    correct += 1
                    total_length += len(h)
                    break
            else:
                incorrect += 1
                to_comp.append((gold, hyp))  # Multiple hypotheses to compare
            self.log_debug(f"For the word {word}: gold is {gold}, hypothesized are: {hyp}")
            hyp_pron_count += len(hyp)
            gold_pron_count += len(gold)
        self.log_debug(
            f"Generated an average of {hyp_pron_count /len(hypothesis_values)} variants "
            f"The gold set had an average of {gold_pron_count/len(hypothesis_values)} variants."
        )
        with mp.Pool(self.num_jobs) as pool:
            gen = pool.starmap(score_g2p, to_comp)
            for (edits, length) in gen:
                total_edits += edits
                total_length += length
        self.wer = 100 * incorrect / (correct + incorrect)
        self.ler = 100 * total_edits / total_length
        self.log_info(f"WER:\t{self.wer:.2f}")
        self.log_info(f"LER:\t{self.ler:.2f}")
        self.log_debug(
            f"Computation of errors for {len(self.g2p_validation_dictionary)} words took {time.time() - begin} seconds"
        )



"""
Aligners
========

"""
from montreal_forced_aligner.alignment.adapting import AdaptingAligner
from montreal_forced_aligner.alignment.base import CorpusAligner
from montreal_forced_aligner.alignment.mixins import AlignMixin
from montreal_forced_aligner.alignment.pretrained import DictionaryTrainer, PretrainedAligner

__all__ = [
    "AdaptingAligner",
    "PretrainedAligner",
    "CorpusAligner",
    "DictionaryTrainer",
    "adapting",
    "base",
    "pretrained",
    "mixins",
    "AlignMixin",
    "multiprocessing",
]



"""Classes for calculating alignments online"""
from __future__ import annotations

import os
import subprocess
import typing

from montreal_forced_aligner.abc import MetaDict
from montreal_forced_aligner.corpus.classes import UtteranceData
from montreal_forced_aligner.data import CtmInterval, MfaArguments
from montreal_forced_aligner.helper import make_safe
from montreal_forced_aligner.textgrid import process_ctm_line
from montreal_forced_aligner.utils import KaldiFunction, thirdparty_binary


class OnlineAlignmentArguments(MfaArguments):
    working_directory: str
    sox_string: str
    utterance_data: UtteranceData
    mfcc_options: MetaDict
    pitch_options: MetaDict
    feature_options: MetaDict
    align_options: MetaDict
    model_path: str
    tree_path: str
    disambig_path: str
    lexicon_fst_path: str
    word_boundary_int_path: str
    reversed_phone_mapping: typing.Dict[int, str]
    optional_silence_phone: str
    silence_words: typing.Set[str]


class OnlineAlignmentFunction(KaldiFunction):
    def __init__(self, args: OnlineAlignmentArguments):
        super(OnlineAlignmentFunction, self).__init__(args)
        self.working_directory = args.working_directory
        self.sox_string = args.sox_string
        self.utterance = args.utterance_data
        self.mfcc_options = args.mfcc_options
        self.pitch_options = args.pitch_options
        self.feature_options = args.feature_options
        self.align_options = args.align_options
        self.model_path = args.model_path
        self.tree_path = args.tree_path
        self.disambig_path = args.disambig_path
        self.lexicon_fst_path = args.lexicon_fst_path
        self.word_boundary_int_path = args.word_boundary_int_path
        self.reversed_phone_mapping = args.reversed_phone_mapping
        self.optional_silence_phone = args.optional_silence_phone
        self.silence_words = args.silence_words

    def cleanup_intervals(self, utterance_name: int, intervals: typing.List[CtmInterval]):
        """
        Clean up phone intervals to remove silence

        Parameters
        ----------
        intervals: list[:class:`~montreal_forced_aligner.data.CtmInterval`]
            Intervals to process

        Returns
        -------
        list[:class:`~montreal_forced_aligner.data.CtmInterval`]
            Cleaned up intervals
        """
        actual_phone_intervals = []
        actual_word_intervals = []
        utterance_name = utterance_name
        utterance_begin = self.utterance.begin
        current_word_begin = None
        words = self.utterance.normalized_text
        words_index = 0
        for interval in intervals:
            interval.begin += utterance_begin
            interval.end += utterance_begin
            phone_label = self.reversed_phone_mapping[int(interval.label)]
            if phone_label == self.optional_silence_phone:
                if words_index < len(words) and words[words_index] in self.silence_words:
                    interval.label = phone_label
                else:
                    interval.label = phone_label
                    actual_phone_intervals.append(interval)
                    continue
            phone, position = phone_label.split("_")
            if position in {"B", "S"}:
                current_word_begin = interval.begin
            if position in {"E", "S"}:
                actual_word_intervals.append(
                    CtmInterval(
                        current_word_begin, interval.end, words[words_index], utterance_name
                    )
                )
                words_index += 1
                current_word_begin = None
            interval.label = phone
            actual_phone_intervals.append(interval)
        return actual_word_intervals, actual_phone_intervals

    def run(self):
        wav_path = os.path.join(self.working_directory, "wav.scp")
        likelihood_path = os.path.join(self.working_directory, "likelihoods.scp")
        feat_path = os.path.join(self.working_directory, "feats.scp")
        cmvn_path = os.path.join(self.working_directory, "cmvn.scp")
        utt2spk_path = os.path.join(self.working_directory, "utt2spk.scp")
        segment_path = os.path.join(self.working_directory, "segments.scp")
        text_int_path = os.path.join(self.working_directory, "text.int")
        lda_mat_path = os.path.join(self.working_directory, "lda.mat")
        fst_path = os.path.join(self.working_directory, "fsts.ark")
        if self.align_options["boost_silence"] != 1.0:
            mdl_string = f"gmm-boost-silence --boost={self.align_options['boost_silence']} {self.align_options['optional_silence_csl']} {self.model_path} - |"

        else:
            mdl_string = self.model_path
        if not os.path.exists(lda_mat_path):
            lda_mat_path = None
        with open(self.log_path, "w") as log_file:
            proc = subprocess.Popen(
                [
                    thirdparty_binary("compile-train-graphs"),
                    f"--read-disambig-syms={self.disambig_path}",
                    self.tree_path,
                    self.model_path,
                    self.lexicon_fst_path,
                    f"ark:{text_int_path}",
                    f"ark:{fst_path}",
                ],
                stderr=log_file,
                encoding="utf8",
                env=os.environ,
            )
            proc.communicate()
            if not os.path.exists(feat_path):
                use_pitch = self.pitch_options.pop("use-pitch")
                mfcc_base_command = [thirdparty_binary("compute-mfcc-feats"), "--verbose=2"]
                for k, v in self.mfcc_options.items():
                    mfcc_base_command.append(f"--{k.replace('_', '-')}={make_safe(v)}")
                mfcc_base_command += ["ark:-", "ark:-"]
                seg_proc = subprocess.Popen(
                    [
                        thirdparty_binary("extract-segments"),
                        f"scp:{wav_path}",
                        segment_path,
                        "ark:-",
                    ],
                    stdout=subprocess.PIPE,
                    stderr=log_file,
                    env=os.environ,
                )
                mfcc_proc = subprocess.Popen(
                    mfcc_base_command,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    stdin=seg_proc.stdout,
                    env=os.environ,
                )
                if use_pitch:
                    pitch_base_command = [
                        thirdparty_binary("compute-and-process-kaldi-pitch-feats"),
                        "--verbose=2",
                    ]
                    for k, v in self.pitch_options.items():
                        pitch_base_command.append(f"--{k.replace('_', '-')}={make_safe(v)}")
                        if k == "delta-pitch":
                            pitch_base_command.append(f"--delta-pitch-noise-stddev={make_safe(v)}")
                    pitch_command = " ".join(pitch_base_command)
                    if os.path.exists(segment_path):
                        segment_command = (
                            f"extract-segments scp:{wav_path} {segment_path} ark:- | "
                        )
                        pitch_input = "ark:-"
                    else:
                        segment_command = ""
                        pitch_input = f"scp:{wav_path}"
                    pitch_feat_string = (
                        f"ark,s,cs:{segment_command}{pitch_command} {pitch_input} ark:- |"
                    )
                    length_tolerance = 2
                    feature_proc = subprocess.Popen(
                        [
                            thirdparty_binary("paste-feats"),
                            f"--length-tolerance={length_tolerance}",
                            "ark:-",
                            pitch_feat_string,
                            "ark:-",
                        ],
                        stdin=mfcc_proc.stdout,
                        env=os.environ,
                        stdout=subprocess.PIPE,
                        stderr=log_file,
                    )
                else:
                    feature_proc = mfcc_proc
                cvmn_proc = subprocess.Popen(
                    [thirdparty_binary("apply-cmvn-sliding"), "--center", "ark:-", "ark:-"],
                    stdin=feature_proc.stdout,
                    env=os.environ,
                    stdout=subprocess.PIPE,
                    stderr=log_file,
                )
                if lda_mat_path is not None:
                    splice_proc = subprocess.Popen(
                        [
                            thirdparty_binary("splice-feats"),
                            f'--left-context={self.feature_options["splice_left_context"]}',
                            f'--right-context={self.feature_options["splice_right_context"]}',
                            "ark:-",
                            "ark:-",
                        ],
                        stdin=cvmn_proc.stdout,
                        env=os.environ,
                        stdout=subprocess.PIPE,
                        stderr=log_file,
                    )
                    transform_proc = subprocess.Popen(
                        [thirdparty_binary("transform-feats"), lda_mat_path, "ark:-", "ark:-"],
                        stdin=splice_proc.stdout,
                        env=os.environ,
                        stdout=subprocess.PIPE,
                        stderr=log_file,
                    )
                elif self.feature_options["uses_deltas"]:
                    transform_proc = subprocess.Popen(
                        [thirdparty_binary("add-deltas"), "ark:-", "ark:-"],
                        stdin=cvmn_proc.stdout,
                        env=os.environ,
                        stdout=subprocess.PIPE,
                        stderr=log_file,
                    )
                # Features done, alignment
                align_proc = subprocess.Popen(
                    [
                        thirdparty_binary("gmm-align-compiled"),
                        f"--transition-scale={self.align_options['transition_scale']}",
                        f"--acoustic-scale={self.align_options['acoustic_scale']}",
                        f"--self-loop-scale={self.align_options['self_loop_scale']}",
                        f"--beam={self.align_options['beam']}",
                        f"--retry-beam={self.align_options['retry_beam']}",
                        "--careful=false",
                        mdl_string,
                        f"ark:{fst_path}",
                        "ark:-",
                        "ark:-",
                        f"ark,t:{likelihood_path}",
                    ],
                    stdout=subprocess.PIPE,
                    stderr=log_file,
                    encoding="utf8",
                    stdin=transform_proc.stdout,
                    env=os.environ,
                )
            else:
                feat_string = f"ark,s,cs:apply-cmvn --utt2spk=ark:{utt2spk_path} scp:{cmvn_path} scp:{feat_path} ark:- |"
                if lda_mat_path is not None:
                    feat_string += f" splice-feats --left-context={self.feature_options['splice_left_context']} --right-context={self.feature_options['splice_right_context']} ark:- ark:- |"
                    feat_string += f" transform-feats {lda_mat_path} ark:- ark:- |"
                    align_proc = subprocess.Popen(
                        [
                            thirdparty_binary("gmm-align-compiled"),
                            f"--transition-scale={self.align_options['transition_scale']}",
                            f"--acoustic-scale={self.align_options['acoustic_scale']}",
                            f"--self-loop-scale={self.align_options['self_loop_scale']}",
                            f"--beam={self.align_options['beam']}",
                            f"--retry-beam={self.align_options['retry_beam']}",
                            "--careful=false",
                            mdl_string,
                            f"ark:{fst_path}",
                            feat_string,
                            "ark:-",
                            f"ark,t:{likelihood_path}",
                        ],
                        stdout=subprocess.PIPE,
                        stderr=log_file,
                        encoding="utf8",
                        env=os.environ,
                    )
            lin_proc = subprocess.Popen(
                [
                    thirdparty_binary("linear-to-nbest"),
                    "ark:-",
                    f"ark:{text_int_path}",
                    "",
                    "",
                    "ark:-",
                ],
                stdin=align_proc.stdout,
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            align_words_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-align-words"),
                    self.word_boundary_int_path,
                    self.model_path,
                    "ark:-",
                    "ark:-",
                ],
                stdin=lin_proc.stdout,
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            phone_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-to-phone-lattice"),
                    self.model_path,
                    "ark:-",
                    "ark:-",
                ],
                stdout=subprocess.PIPE,
                stdin=align_words_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            nbest_proc = subprocess.Popen(
                [
                    thirdparty_binary("nbest-to-ctm"),
                    "--print-args=false",
                    f"--frame-shift={round(self.feature_options['frame_shift']/1000,4)}",
                    "ark:-",
                    "-",
                ],
                stdin=phone_proc.stdout,
                stderr=log_file,
                stdout=subprocess.PIPE,
                env=os.environ,
                encoding="utf8",
            )
            intervals = []
            log_likelihood = None
            for line in nbest_proc.stdout:
                line = line.strip()
                if not line:
                    continue
                try:
                    interval = process_ctm_line(line)
                    intervals.append(interval)
                except ValueError:
                    pass
            nbest_proc.wait()
            with open(likelihood_path, "r", encoding="utf8") as f:
                try:
                    log_likelihood = float(f.read().split()[-1])
                except ValueError:
                    pass
        return *self.cleanup_intervals(0, intervals), log_likelihood
